import logging
from typing import List

import numpy as np
import pandas as pd

from autorag.nodes.promptmaker.base import BasePromptMaker
from autorag.utils import result_to_dataframe

logger = logging.getLogger("AutoRAG")


class LongContextReorder(BasePromptMaker):
	@result_to_dataframe(["prompts"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		query, retrieved_contents, prompt = self.cast_to_run(
			previous_result, *args, **kwargs
		)
		assert (
			"retrieve_scores" in previous_result.columns
		), "previous_result must have retrieve_scores column."
		retrieve_scores = previous_result["retrieve_scores"].tolist()
		return self._pure(prompt, query, retrieved_contents, retrieve_scores)

	def _pure(
		self,
		prompt: str,
		queries: List[str],
		retrieved_contents: List[List[str]],
		retrieve_scores: List[List[float]],
	) -> List[str]:
		"""
		Models struggle to access significant details found
		in the center of extended contexts. A study
		(https://arxiv.org/abs/2307.03172) observed that the best
		performance typically arises when crucial data is positioned
		at the start or conclusion of the input context. Additionally,
		as the input context lengthens, performance drops notably, even
		in models designed for long contexts."

		.. Code:: yaml
		nodes:
		- node_type: prompt_maker
		  modules:
		  - module_type: long_context_reorder
		    prompt: [Answer this question: {query} \n\n {retrieved_contents},
		    Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

		:param prompt: A prompt string.
		:param queries: List of query strings.
		:param retrieved_contents: List of retrieved contents.
		:param retrieve_scores: List of `retrieve scores`.
		:return: Prompts that are made by long context reorder.
		"""

		def long_context_reorder_row(
			_prompt: str,
			_query: str,
			_retrieved_contents: List[str],
			_retrieve_scores: List[float],
		) -> str:
			if isinstance(_retrieved_contents, np.ndarray):
				_retrieved_contents = _retrieved_contents.tolist()
			if not len(_retrieved_contents) == len(_retrieve_scores):
				logger.info("If you use a summarizer, the reorder will not proceed.")
				return _prompt.format(
					query=_query, retrieved_contents="\n\n".join(_retrieved_contents)
				)
			content_scores = list(zip(_retrieved_contents, _retrieve_scores))
			sorted_content_scores = sorted(
				content_scores, key=lambda x: x[1], reverse=True
			)
			content_result, score_result = zip(*sorted_content_scores)
			_retrieved_contents.append(content_result[0])
			contents_str = "\n\n".join(_retrieved_contents)
			return _prompt.format(query=_query, retrieved_contents=contents_str)

		return list(
			map(
				lambda x: long_context_reorder_row(prompt, x[0], x[1], x[2]),
				zip(queries, retrieved_contents, retrieve_scores),
			)
		)



import os
import pathlib
from typing import List, Dict, Union

import pandas as pd

from autorag.evaluation import evaluate_generation
from autorag.evaluation.util import cast_metrics
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.utils.util import to_list


def run_generator_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run evaluation and select the best module among generator node results.
	And save the results and summary to generator node directory.

	:param modules: Generator modules to run.
	:param module_params: Generator module parameters.
	    Including node parameters, which is used for every module in this node.
	:param previous_result: Previous result dataframe.
	    Could be prompt maker node's result.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for generator node.
	:return: The best result dataframe.
	    It contains previous result columns and generator node's result columns.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent
	node_dir = os.path.join(node_line_dir, "generator")  # node name
	if not os.path.exists(node_dir):
		os.makedirs(node_dir)
	qa_data = pd.read_parquet(
		os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
	)
	if "generation_gt" not in qa_data.columns:
		raise ValueError("You must have 'generation_gt' column in qa.parquet.")

	results, execution_times = zip(
		*map(
			lambda x: measure_speed(
				x[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**x[1],
			),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# get average token usage
	token_usages = list(map(lambda x: x["generated_tokens"].apply(len).mean(), results))

	# make rows to metric_inputs
	generation_gt = to_list(qa_data["generation_gt"].tolist())

	metric_inputs = [MetricInput(generation_gt=gen_gt) for gen_gt in generation_gt]

	metric_names, metric_params = cast_metrics(strategies.get("metrics"))
	if metric_names is None or len(metric_names) <= 0:
		raise ValueError("You must at least one metrics for generator evaluation.")
	results = list(
		map(
			lambda result: evaluate_generator_node(
				result, metric_inputs, strategies.get("metrics")
			),
			results,
		)
	)

	# save results to folder
	filepaths = list(
		map(lambda x: os.path.join(node_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(
		map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths))
	)  # execute save to parquet
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
			"average_output_token": token_usages,
			**{
				metric: list(map(lambda x: x[metric].mean(), results))
				for metric in metric_names
			},
		}
	)

	# filter by strategies
	if strategies.get("speed_threshold") is not None:
		results, filenames = filter_by_threshold(
			results, average_times, strategies["speed_threshold"], filenames
		)
	if strategies.get("token_threshold") is not None:
		results, filenames = filter_by_threshold(
			results, token_usages, strategies["token_threshold"], filenames
		)
	selected_result, selected_filename = select_best(
		results, metric_names, filenames, strategies.get("strategy", "mean")
	)
	best_result = pd.concat([previous_result, selected_result], axis=1)

	# add 'is_best' column at summary file
	summary_df["is_best"] = summary_df["filename"] == selected_filename

	# save files
	summary_df.to_csv(os.path.join(node_dir, "summary.csv"), index=False)
	best_result.to_parquet(
		os.path.join(
			node_dir, f"best_{os.path.splitext(selected_filename)[0]}.parquet"
		),
		index=False,
	)
	return best_result


def evaluate_generator_node(
	result_df: pd.DataFrame,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
):
	@evaluate_generation(metric_inputs=metric_inputs, metrics=metrics)
	def evaluate_generation_module(df: pd.DataFrame):
		return (
			df["generated_texts"].tolist(),
			df["generated_tokens"].tolist(),
			df["generated_log_probs"].tolist(),
		)

	return evaluate_generation_module(result_df)



import os
import pathlib
from typing import List, Dict, Union

import pandas as pd

from autorag.evaluation import evaluate_retrieval
from autorag.evaluation.retrieval import RETRIEVAL_METRIC_FUNC_DICT
from autorag.nodes.retrieval.run_util import save_and_summary, find_best
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed
from autorag.utils.util import apply_recursive, to_list


def run_semantic_retrieval_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run the semantic retrieval node.

	:param modules: Retrieval modules to run.
	:param module_params: Retrieval module parameters.
	:param previous_result: Previous result dataframe.
	    Could be query expansion's best result or qa data.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for retrieval node.
	:return: The best result dataframe.
	    It contains previous result columns and retrieval node's result columns.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent
	qa_df = pd.read_parquet(
		os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
	)
	retrieval_gt = qa_df["retrieval_gt"].tolist()
	retrieval_gt = apply_recursive(lambda x: str(x), to_list(retrieval_gt))
	# make rows to metric_inputs
	metric_inputs = [
		MetricInput(retrieval_gt=ret_gt, query=query, generation_gt=gen_gt)
		for ret_gt, query, gen_gt in zip(
			retrieval_gt, qa_df["query"].tolist(), qa_df["generation_gt"].tolist()
		)
	]

	save_dir = os.path.join(node_line_dir, "semantic_retrieval")
	if not os.path.exists(save_dir):
		os.makedirs(save_dir)

	# Run the modules
	semantic_results, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(modules, module_params),
		)
	)
	semantic_times = list(map(lambda x: x / len(semantic_results[0]), execution_times))

	# run metrics
	if strategies.get("metrics") is None:
		raise ValueError("You must at least one metrics for retrieval evaluation.")
	semantic_results = list(
		map(
			lambda x: evaluate_semantic_retrieval_node(
				x,
				metric_inputs,
				strategies.get("metrics"),
			),
			semantic_results,
		)
	)

	semantic_summary_df = save_and_summary(
		modules,
		module_params,
		semantic_results,
		semantic_times,
		0,
		save_dir,
		strategies,
	)
	semantic_selected_result, semantic_selected_filename = find_best(
		semantic_results,
		semantic_times,
		semantic_summary_df["filename"].tolist(),
		strategies,
	)
	semantic_summary_df["is_best"] = (
		semantic_summary_df["filename"] == semantic_selected_filename
	)
	previous_result.drop(
		columns=list(RETRIEVAL_METRIC_FUNC_DICT.keys()), inplace=True, errors="ignore"
	)
	semantic_selected_result.rename(
		columns={
			"retrieved_contents": "retrieved_contents_semantic",
			"retrieved_ids": "retrieved_ids_semantic",
			"retrieve_scores": "retrieve_scores_semantic",
		},
		inplace=True,
	)
	best_result = pd.concat([previous_result, semantic_selected_result], axis=1)
	best_result.to_parquet(
		os.path.join(
			save_dir, f"best_{os.path.splitext(semantic_selected_filename)[0]}.parquet"
		),
		index=False,
	)
	semantic_summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
	return best_result  # The result will be with _semantic suffix


def evaluate_semantic_retrieval_node(
	result_df: pd.DataFrame,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
) -> pd.DataFrame:
	"""
	Evaluate retrieval node from retrieval node result dataframe.

	:param result_df: The result dataframe from a retrieval node.
	:param metric_inputs: List of metric input schema for AutoRAG.
	:param metrics: Metric list from input strategies.
	:return: Return result_df with metrics columns.
	    The columns will be 'retrieved_contents_semantic', 'retrieved_ids_semantic', 'retrieve_scores_semantic', and metric names.
	"""

	@evaluate_retrieval(
		metric_inputs=metric_inputs,
		metrics=metrics,
	)
	def evaluate_this_module(df: pd.DataFrame):
		return (
			df["retrieved_contents_semantic"].tolist(),
			df["retrieved_ids_semantic"].tolist(),
			df["retrieve_scores_semantic"].tolist(),
		)

	return evaluate_this_module(result_df)



import logging
import os
import shutil
from typing import Optional

from autorag.data.parse.run import run_parser
from autorag.data.utils.util import load_yaml, get_param_combinations

logger = logging.getLogger("AutoRAG")


class Parser:
	def __init__(self, data_path_glob: str, project_dir: Optional[str] = None):
		self.data_path_glob = data_path_glob
		self.project_dir = project_dir if project_dir is not None else os.getcwd()

	def start_parsing(self, yaml_path: str, all_files: bool = False):
		if not os.path.exists(self.project_dir):
			os.makedirs(self.project_dir)

		# copy yaml file to project directory
		shutil.copy(yaml_path, os.path.join(self.project_dir, "parse_config.yaml"))

		# load yaml file
		modules = load_yaml(yaml_path)

		input_modules, input_params = get_param_combinations(modules)

		logger.info("Parsing Start...")
		run_parser(
			modules=input_modules,
			module_params=input_params,
			data_path_glob=self.data_path_glob,
			project_dir=self.project_dir,
			all_files=all_files,
		)
		logger.info("Parsing Done!")



import abc
import logging
from pathlib import Path
from typing import Union

import pandas as pd

from autorag.schema.base import BaseModule
from autorag.utils import validate_qa_dataset
from autorag.utils.cast import cast_retrieve_infos

logger = logging.getLogger("AutoRAG")


class BasePassageFilter(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		logger.info(f"Initialize passage filter node - {self.__class__.__name__}")

	def __del__(self):
		logger.info(f"Prompt maker node - {self.__class__.__name__} module is deleted.")

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(
			f"Running passage filter node - {self.__class__.__name__} module..."
		)
		validate_qa_dataset(previous_result)

		# find queries columns
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		queries = previous_result["query"].tolist()

		retrieve_infos = cast_retrieve_infos(previous_result)
		return (
			queries,
			retrieve_infos["retrieved_contents"],
			retrieve_infos["retrieve_scores"],
			retrieve_infos["retrieved_ids"],
		)



# This is v2 version, the next version of data creation
# The legacy (v1) version will be deprecated on AutoRAG version 0.3
# The legacy (v1) version and new v2 data creation is not compatible with each other



import os.path
import pathlib
from typing import List, Dict

import pandas as pd

from autorag.evaluation.metric import (
	retrieval_token_recall,
	retrieval_token_precision,
	retrieval_token_f1,
)
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.utils.util import fetch_contents


def run_passage_compressor_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run evaluation and select the best module among passage compressor modules.

	:param modules: Passage compressor modules to run.
	:param module_params: Passage compressor module parameters.
	:param previous_result: Previous result dataframe.
	    Could be retrieval, reranker modules result.
	    It means it must contain 'query', 'retrieved_contents', 'retrieved_ids', 'retrieve_scores' columns.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for passage compressor node.
	    In this node, we use
	    You can skip evaluation when you use only one module and a module parameter.
	:return: The best result dataframe with previous result columns.
	    This node will replace 'retrieved_contents' to compressed passages, so its length will be one.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent
	data_dir = os.path.join(project_dir, "data")
	save_dir = os.path.join(node_line_dir, "passage_compressor")
	if not os.path.exists(save_dir):
		os.makedirs(save_dir)

	# make retrieval contents gt
	qa_data = pd.read_parquet(os.path.join(data_dir, "qa.parquet"), engine="pyarrow")
	corpus_data = pd.read_parquet(
		os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
	)
	# check qa_data have retrieval_gt
	assert all(
		len(x[0]) > 0 for x in qa_data["retrieval_gt"].tolist()
	), "Can't use passage compressor if you don't have retrieval gt values in QA dataset."

	# run modules
	results, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(modules, module_params),
		)
	)
	results = list(results)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	retrieval_gt_contents = list(
		map(lambda x: fetch_contents(corpus_data, x), qa_data["retrieval_gt"].tolist())
	)

	metric_inputs = [
		MetricInput(retrieval_gt_contents=ret_cont_gt)
		for ret_cont_gt in retrieval_gt_contents
	]

	# run metrics before filtering
	if strategies.get("metrics") is None:
		raise ValueError(
			"You must at least one metrics for retrieval contents evaluation."
			"It can be 'retrieval_token_f1', 'retrieval_token_precision', 'retrieval_token_recall'."
		)
	results = list(
		map(
			lambda x: evaluate_passage_compressor_node(
				x, metric_inputs, strategies.get("metrics")
			),
			results,
		)
	)

	# save results to folder
	filepaths = list(
		map(lambda x: os.path.join(save_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(
		map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths))
	)  # execute save to parquet
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	# make summary file
	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
			**{
				f"passage_compressor_{metric}": list(
					map(lambda result: result[metric].mean(), results)
				)
				for metric in strategies.get("metrics")
			},
		}
	)

	# filter by strategies
	if strategies.get("speed_threshold") is not None:
		results, filenames = filter_by_threshold(
			results, average_times, strategies["speed_threshold"], filenames
		)
	selected_result, selected_filename = select_best(
		results,
		strategies.get("metrics"),
		filenames,
		strategies.get("strategy", "mean"),
	)
	new_retrieved_contents = selected_result["retrieved_contents"]
	previous_result["retrieved_contents"] = new_retrieved_contents
	selected_result = selected_result.drop(columns=["retrieved_contents"])
	best_result = pd.concat([previous_result, selected_result], axis=1)

	# add 'is_best' column to summary file
	summary_df["is_best"] = summary_df["filename"] == selected_filename

	# add prefix 'passage_compressor' to best_result columns
	best_result = best_result.rename(
		columns={
			metric_name: f"passage_compressor_{metric_name}"
			for metric_name in strategies.get("metrics")
		}
	)

	# save the result files
	best_result.to_parquet(
		os.path.join(
			save_dir, f"best_{os.path.splitext(selected_filename)[0]}.parquet"
		),
		index=False,
	)
	summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
	return best_result


def evaluate_passage_compressor_node(
	result_df: pd.DataFrame, metric_inputs: List[MetricInput], metrics: List[str]
):
	metric_funcs = {
		retrieval_token_recall.__name__: retrieval_token_recall,
		retrieval_token_precision.__name__: retrieval_token_precision,
		retrieval_token_f1.__name__: retrieval_token_f1,
	}
	for metric_input, generated_text in zip(
		metric_inputs, result_df["retrieved_contents"].tolist()
	):
		metric_input.retrieved_contents = generated_text
	metrics = list(filter(lambda x: x in metric_funcs.keys(), metrics))
	if len(metrics) <= 0:
		raise ValueError(f"metrics must be one of {metric_funcs.keys()}")
	metrics_scores = dict(
		map(
			lambda metric: (
				metric,
				metric_funcs[metric](
					metric_inputs=metric_inputs,
				),
			),
			metrics,
		)
	)
	result_df = pd.concat([result_df, pd.DataFrame(metrics_scores)], axis=1)
	return result_df



import functools
import logging
from typing import Tuple, List, Dict, Any

import pandas as pd

from autorag.embedding.base import EmbeddingModel
from autorag.data import chunk_modules, sentence_splitter_modules
from autorag.utils import result_to_dataframe

logger = logging.getLogger("AutoRAG")


def chunker_node(func):
	@functools.wraps(func)
	@result_to_dataframe(["doc_id", "contents", "path", "start_end_idx", "metadata"])
	def wrapper(
		parsed_result: pd.DataFrame, chunk_method: str, **kwargs
	) -> Tuple[
		List[str], List[str], List[str], List[Tuple[int, int]], List[Dict[str, Any]]
	]:
		logger.info(f"Running chunker - {func.__name__} module...")

		# get texts from parsed_result
		texts = parsed_result["texts"].tolist()

		# get filenames from parsed_result when 'add_file_name' is setting
		file_name_language = kwargs.pop("add_file_name", None)
		metadata_list = make_metadata_list(parsed_result)

		# run chunk module
		if func.__name__ in ["llama_index_chunk", "langchain_chunk"]:
			chunk_instance = __get_chunk_instance(
				func.__name__, chunk_method.lower(), **kwargs
			)
			result = func(
				texts=texts,
				chunker=chunk_instance,
				file_name_language=file_name_language,
				metadata_list=metadata_list,
			)
			del chunk_instance
			return result
		else:
			raise ValueError(f"Unsupported module_type: {func.__name__}")

	return wrapper


def make_metadata_list(parsed_result: pd.DataFrame) -> List[Dict[str, str]]:
	metadata_list = [{} for _ in range(len(parsed_result["texts"]))]

	def _make_metadata_pure(
		lst: List[str], key: str, metadata_lst: List[Dict[str, str]]
	):
		for value, metadata in zip(lst, metadata_lst):
			metadata[key] = value

	for column in ["page", "last_modified_datetime", "path"]:
		if column in parsed_result.columns:
			_make_metadata_pure(parsed_result[column].tolist(), column, metadata_list)
	return metadata_list


def __get_chunk_instance(module_type: str, chunk_method: str, **kwargs):
	# Add sentence_splitter to kwargs
	sentence_available_methods = [
		"semantic_llama_index",
		"semanticdoublemerging",
		"sentencewindow",
	]
	if chunk_method in sentence_available_methods:
		# llama index default sentence_splitter is 'nltk -PunktSentenceTokenizer'
		if "sentence_splitter" in kwargs.keys():
			sentence_splitter_str = kwargs.pop("sentence_splitter")
			sentence_splitter_func = sentence_splitter_modules[sentence_splitter_str]()
			kwargs.update({"sentence_splitter": sentence_splitter_func})

	def get_embedding_model(_embed_model_str: str, _module_type: str):
		if _embed_model_str == "openai":
			if _module_type == "langchain_chunk":
				_embed_model_str = "openai_langchain"
		return EmbeddingModel.load(_embed_model_str)()

	# Add embed_model to kwargs
	embedding_available_methods = ["semantic_llama_index", "semantic_langchain"]
	if chunk_method in embedding_available_methods:
		# there is no default embed_model, so we have to get it parameter and add it.
		if "embed_model" not in kwargs.keys():
			raise ValueError(f"embed_model is required for {chunk_method} method.")
		embed_model_str = kwargs.pop("embed_model")
		embed_model = get_embedding_model(embed_model_str, module_type)
		if chunk_method == "semantic_llama_index":
			kwargs.update({"embed_model": embed_model})
		elif chunk_method == "semantic_langchain":
			kwargs.update({"embeddings": embed_model})

	return chunk_modules[chunk_method](**kwargs)


def add_file_name(
	file_name_language: str, file_names: List[str], chunk_texts: List[str]
) -> List[str]:
	if file_name_language == "en":
		return list(
			map(
				lambda x: f"file_name: {x[1]}\n contents: {x[0]}",
				zip(chunk_texts, file_names),
			)
		)
	elif file_name_language == "ko":
		return list(
			map(
				lambda x: f"파일 제목: {x[1]}\n 내용: {x[0]}",
				zip(chunk_texts, file_names),
			)
		)
	elif file_name_language == "ja":
		return list(
			map(
				lambda x: f"ファイル名: {x[1]}\n 内容: {x[0]}",
				zip(chunk_texts, file_names),
			)
		)
	else:
		raise ValueError(
			f"Unsupported file_name_language: {file_name_language}. Choose from 'en' ,'ko' or 'ja."
		)



import ast
import logging
import os
from typing import Dict, List

import matplotlib.pyplot as plt
import pandas as pd
import panel as pn
import seaborn as sns
import yaml
from bokeh.models import NumberFormatter, BooleanFormatter

from autorag.utils.util import dict_to_markdown, dict_to_markdown_table

pn.extension(
	"terminal",
	"tabulator",
	"mathjax",
	"ipywidgets",
	console_output="disable",
	sizing_mode="stretch_width",
	css_files=[
		"https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css"
	],
)
logger = logging.getLogger("AutoRAG")


def find_node_dir(trial_dir: str) -> List[str]:
	trial_summary_df = pd.read_csv(os.path.join(trial_dir, "summary.csv"))
	result_paths = []
	for idx, row in trial_summary_df.iterrows():
		node_line_name = row["node_line_name"]
		node_type = row["node_type"]
		result_paths.append(os.path.join(trial_dir, node_line_name, node_type))
	return result_paths


def get_metric_values(node_summary_df: pd.DataFrame) -> Dict:
	non_metric_column_names = [
		"filename",
		"module_name",
		"module_params",
		"execution_time",
		"average_output_token",
		"is_best",
	]
	best_row = node_summary_df.loc[node_summary_df["is_best"]].drop(
		columns=non_metric_column_names, errors="ignore"
	)
	assert len(best_row) == 1, "The best module must be only one."
	return best_row.iloc[0].to_dict()


def make_trial_summary_md(trial_dir):
	markdown_text = f"""# Trial Result Summary
- Trial Directory : {trial_dir}

"""
	node_dirs = find_node_dir(trial_dir)
	for node_dir in node_dirs:
		node_summary_filepath = os.path.join(node_dir, "summary.csv")
		node_type = os.path.basename(node_dir)
		node_summary_df = pd.read_csv(node_summary_filepath)
		best_row = node_summary_df.loc[node_summary_df["is_best"]].iloc[0]
		metric_dict = get_metric_values(node_summary_df)
		markdown_text += f"""---

## {node_type} best module

### Module Name

{best_row['module_name']}

### Module Params

{dict_to_markdown(ast.literal_eval(best_row['module_params']), level=3)}

### Metric Values

{dict_to_markdown_table(metric_dict, key_column_name='metric_name', value_column_name='metric_value')}

"""

	return markdown_text


def node_view(node_dir: str):
	non_metric_column_names = [
		"filename",
		"module_name",
		"module_params",
		"execution_time",
		"average_output_token",
		"is_best",
	]
	summary_df = pd.read_csv(os.path.join(node_dir, "summary.csv"))
	bokeh_formatters = {
		"float": NumberFormatter(format="0.000"),
		"bool": BooleanFormatter(),
	}
	first_df = pd.read_parquet(os.path.join(node_dir, "0.parquet"), engine="pyarrow")

	each_module_df_widget = pn.widgets.Tabulator(
		pd.DataFrame(columns=first_df.columns),
		name="Module DataFrame",
		formatters=bokeh_formatters,
		pagination="local",
		page_size=20,
		widths=150,
	)

	def change_module_widget(event):
		if event.column == "detail":
			filename = summary_df["filename"].iloc[event.row]
			filepath = os.path.join(node_dir, filename)
			each_module_df = pd.read_parquet(filepath, engine="pyarrow")
			each_module_df_widget.value = each_module_df

	df_widget = pn.widgets.Tabulator(
		summary_df,
		name="Summary DataFrame",
		formatters=bokeh_formatters,
		buttons={"detail": '<i class="fa fa-eye"></i>'},
		widths=150,
	)
	df_widget.on_click(change_module_widget)

	try:
		fig, ax = plt.subplots(figsize=(10, 5))
		metric_df = summary_df.drop(columns=non_metric_column_names, errors="ignore")
		sns.stripplot(data=metric_df, ax=ax)
		strip_plot_pane = pn.pane.Matplotlib(fig, tight=True)

		fig2, ax2 = plt.subplots(figsize=(10, 5))
		sns.boxplot(data=metric_df, ax=ax2)
		box_plot_pane = pn.pane.Matplotlib(fig2, tight=True)
		plot_pane = pn.Row(strip_plot_pane, box_plot_pane)

		layout = pn.Column(
			"## Summary distribution plot",
			plot_pane,
			"## Summary DataFrame",
			df_widget,
			"## Module Result DataFrame",
			each_module_df_widget,
		)
	except Exception as e:
		logger.error(f"Skipping make boxplot and stripplot with error {e}")
		layout = pn.Column("## Summary DataFrame", df_widget)
	layout.servable()
	return layout


CSS = """
div.card-margin:nth-child(1) {
    max-height: 300px;
}
div.card-margin:nth-child(2) {
    max-height: 400px;
}
"""


def yaml_to_markdown(yaml_filepath):
	markdown_content = ""
	with open(yaml_filepath, "r", encoding="utf-8") as file:
		try:
			content = yaml.safe_load(file)
			markdown_content += f"## {os.path.basename(yaml_filepath)}\n```yaml\n{yaml.safe_dump(content, allow_unicode=True)}\n```\n\n"
		except yaml.YAMLError as exc:
			print(f"Error in {yaml_filepath}: {exc}")
	return markdown_content


def run(trial_dir: str, port: int = 7690):
	trial_summary_md = make_trial_summary_md(trial_dir=trial_dir)
	trial_summary_tab = pn.pane.Markdown(trial_summary_md, sizing_mode="stretch_width")

	node_views = [
		(str(os.path.basename(node_dir)), node_view(node_dir))
		for node_dir in find_node_dir(trial_dir)
	]

	yaml_file_markdown = yaml_to_markdown(os.path.join(trial_dir, "config.yaml"))

	yaml_file = pn.pane.Markdown(yaml_file_markdown, sizing_mode="stretch_width")

	tabs = pn.Tabs(
		("Summary", trial_summary_tab),
		*node_views,
		("Used YAML file", yaml_file),
		dynamic=True,
	)

	template = pn.template.FastListTemplate(
		site="AutoRAG", title="Dashboard", main=[tabs], raw_css=[CSS]
	).servable()
	template.show(port=port)



import glob
import json
import logging
import os
import shutil
from datetime import datetime
from itertools import chain
from typing import List, Dict, Optional

import pandas as pd
import yaml

from autorag.node_line import run_node_line
from autorag.nodes.retrieval.base import get_bm25_pkl_name
from autorag.nodes.lexicalretrieval.bm25 import bm25_ingest
from autorag.nodes.semanticretrieval.vectordb import (
	vectordb_ingest_api,
	filter_exist_ids,
	filter_exist_ids_from_retrieval_gt,
	vectordb_ingest_huggingface,
)
from autorag.schema import Node
from autorag.schema.node import (
	module_type_exists,
	extract_values_from_nodes,
	extract_values_from_nodes_strategy,
)
from autorag.utils import (
	cast_qa_dataset,
	cast_corpus_dataset,
	validate_qa_from_corpus_dataset,
)
from autorag.utils.util import (
	load_summary_file,
	explode,
	load_yaml_config,
	get_event_loop,
)
from autorag.vectordb import load_all_vectordb_from_yaml

logger = logging.getLogger("AutoRAG")

ascii_art = """
                _        _____            _____
     /\        | |      |  __ \     /\   / ____|
    /  \  _   _| |_ ___ | |__) |   /  \ | |  __
   / /\ \| | | | __/ _ \|  _  /   / /\ \| | |_ |
  / ____ \ |_| | || (_) | | \ \  / ____ \ |__| |
 /_/    \_\__,_|\__\___/|_|  \_\/_/    \_\_____|

"""


class Evaluator:
	def __init__(
		self,
		qa_data_path: str,
		corpus_data_path: str,
		project_dir: Optional[str] = None,
	):
		"""
		Initialize an Evaluator object.

		:param qa_data_path: The path to the QA dataset.
		    Must be parquet file.
		:param corpus_data_path: The path to the corpus dataset.
		    Must be parquet file.
		:param project_dir: The path to the project directory.
		    Default is the current directory.
		"""
		# validate data paths
		if not os.path.exists(qa_data_path):
			raise ValueError(f"QA data path {qa_data_path} does not exist.")
		if not os.path.exists(corpus_data_path):
			raise ValueError(f"Corpus data path {corpus_data_path} does not exist.")
		if not qa_data_path.endswith(".parquet"):
			raise ValueError(f"QA data path {qa_data_path} is not a parquet file.")
		if not corpus_data_path.endswith(".parquet"):
			raise ValueError(
				f"Corpus data path {corpus_data_path} is not a parquet file."
			)
		self.qa_data_path = qa_data_path
		self.corpus_data_path = corpus_data_path
		self.qa_data = pd.read_parquet(qa_data_path, engine="pyarrow")
		self.corpus_data = pd.read_parquet(corpus_data_path, engine="pyarrow")
		self.qa_data = cast_qa_dataset(self.qa_data)
		self.corpus_data = cast_corpus_dataset(self.corpus_data)
		self.project_dir = project_dir if project_dir is not None else os.getcwd()
		if not os.path.exists(self.project_dir):
			os.makedirs(self.project_dir)

		validate_qa_from_corpus_dataset(self.qa_data, self.corpus_data)

		# copy dataset to the project directory
		if not os.path.exists(os.path.join(self.project_dir, "data")):
			os.makedirs(os.path.join(self.project_dir, "data"))
		qa_path_in_project = os.path.join(self.project_dir, "data", "qa.parquet")
		if not os.path.exists(qa_path_in_project):
			self.qa_data.to_parquet(qa_path_in_project, index=False)
		corpus_path_in_project = os.path.join(
			self.project_dir, "data", "corpus.parquet"
		)
		if not os.path.exists(corpus_path_in_project):
			self.corpus_data.to_parquet(corpus_path_in_project, index=False)

	def start_trial(
		self,
		yaml_path: str,
		skip_validation: bool = False,
		full_ingest: bool = True,
	):
		"""
		Start AutoRAG trial.
		The trial means one experiment to optimize the RAG pipeline.
		It consists of ingesting corpus data, running all nodes and modules, evaluating and finding the optimal modules.

		:param yaml_path: The config YAML path
		:param skip_validation: If True, it skips the validation step.
			The validation step checks the input config YAML file is well formatted,
			and there is any problem with the system settings.
			Default is False.
		:param full_ingest: If True, it checks the whole corpus data from corpus.parquet that exists in the Vector DB.
			If your corpus is huge and don't want to check the whole vector DB, please set it to False.
		:return: None
		"""
		# Make Resources directory
		os.makedirs(os.path.join(self.project_dir, "resources"), exist_ok=True)

		if not skip_validation:
			logger.info(ascii_art)
			logger.info(
				"Start Validation input data and config YAML file first. "
				"If you want to skip this, put the --skip_validation flag or "
				"`skip_validation` at the start_trial function."
			)
			from autorag.validator import Validator  # resolve circular import

			validator = Validator(
				qa_data_path=self.qa_data_path, corpus_data_path=self.corpus_data_path
			)
			validator.validate(yaml_path)

		os.environ["PROJECT_DIR"] = self.project_dir

		trial_name = self.__get_new_trial_name()
		self.__make_trial_dir(trial_name)

		# copy YAML file to the trial directory
		shutil.copy(
			yaml_path, os.path.join(self.project_dir, trial_name, "config.yaml")
		)
		yaml_dict = load_yaml_config(yaml_path)
		vectordb = yaml_dict.get("vectordb", [])

		vectordb_config_path = os.path.join(
			self.project_dir, "resources", "vectordb.yaml"
		)
		with open(vectordb_config_path, "w") as f:
			yaml.safe_dump({"vectordb": vectordb}, f)

		node_lines = self._load_node_lines(yaml_path)
		self.__ingest_bm25_full(node_lines)

		# Ingest VectorDB corpus
		if any(
			list(
				map(
					lambda nodes: module_type_exists(nodes, "vectordb"),
					node_lines.values(),
				)
			)
		):
			vectordb_list = load_all_vectordb_from_yaml(yaml_path, self.project_dir)
			for vectordb in vectordb_list:
				loop = get_event_loop()
				target_corpus = loop.run_until_complete(
					self.__get_ingest_target_corpus(vectordb, full_ingest)
				)
				if vectordb.embedding.__class__.class_name() == "HuggingFaceEmbedding":
					vectordb_ingest_huggingface(vectordb, target_corpus)
				else:
					# API Ingest Method
					loop = get_event_loop()
					loop.run_until_complete(
						vectordb_ingest_api(vectordb, target_corpus)
					)

		trial_summary_df = pd.DataFrame(
			columns=[
				"node_line_name",
				"node_type",
				"best_module_filename",
				"best_module_name",
				"best_module_params",
				"best_execution_time",
			]
		)

		for i, (node_line_name, node_line) in enumerate(node_lines.items()):
			node_line_dir = os.path.join(self.project_dir, trial_name, node_line_name)
			os.makedirs(node_line_dir, exist_ok=False)
			if i == 0:
				previous_result = self.qa_data
			logger.info(f"Running node line {node_line_name}...")
			previous_result = run_node_line(
				node_line,
				node_line_dir,
				previous_result,
			)

			trial_summary_df = self._append_node_line_summary(
				node_line_name, node_line_dir, trial_summary_df
			)

		trial_summary_df.to_csv(
			os.path.join(self.project_dir, trial_name, "summary.csv"), index=False
		)

		logger.info("Evaluation complete.")

	def __ingest_bm25_full(self, node_lines: Dict[str, List[Node]]):
		if any(
			list(
				map(
					lambda nodes: module_type_exists(nodes, "bm25"), node_lines.values()
				)
			)
		):
			logger.info("Embedding BM25 corpus...")
			bm25_tokenizer_list = list(
				chain.from_iterable(
					map(
						lambda nodes: self._find_bm25_tokenizer(nodes),
						node_lines.values(),
					)
				)
			)

			if len(bm25_tokenizer_list) == 0:
				bm25_tokenizer_list = ["porter_stemmer"]
			for bm25_tokenizer in bm25_tokenizer_list:
				bm25_dir = os.path.join(
					self.project_dir, "resources", get_bm25_pkl_name(bm25_tokenizer)
				)
				if not os.path.exists(os.path.dirname(bm25_dir)):
					os.makedirs(os.path.dirname(bm25_dir))
				# ingest because bm25 supports update new corpus data
				bm25_ingest(bm25_dir, self.corpus_data, bm25_tokenizer=bm25_tokenizer)
			logger.info("BM25 corpus embedding complete.")

	def __get_new_trial_name(self) -> str:
		trial_json_path = os.path.join(self.project_dir, "trial.json")
		if not os.path.exists(trial_json_path):
			return "0"
		with open(trial_json_path, "r") as f:
			trial_json = json.load(f)
		return str(int(trial_json[-1]["trial_name"]) + 1)

	def __make_trial_dir(self, trial_name: str):
		trial_json_path = os.path.join(self.project_dir, "trial.json")
		if os.path.exists(trial_json_path):
			with open(trial_json_path, "r") as f:
				trial_json = json.load(f)
		else:
			trial_json = []

		trial_json.append(
			{
				"trial_name": trial_name,
				"start_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
			}
		)
		os.makedirs(os.path.join(self.project_dir, trial_name))
		with open(trial_json_path, "w") as f:
			json.dump(trial_json, f, indent=4)

	@staticmethod
	def _load_node_lines(yaml_path: str) -> Dict[str, List[Node]]:
		yaml_dict = load_yaml_config(yaml_path)
		node_lines = yaml_dict["node_lines"]
		node_line_dict = {}
		for node_line in node_lines:
			node_line_dict[node_line["node_line_name"]] = list(
				map(lambda node: Node.from_dict(node), node_line["nodes"])
			)
		return node_line_dict

	def restart_trial(self, trial_path: str):
		logger.info(ascii_art)
		os.environ["PROJECT_DIR"] = self.project_dir
		# Check if trial_path exists
		if not os.path.exists(trial_path):
			raise ValueError(f"Trial path {trial_path} does not exist.")
		# Check if trial is completed
		if os.path.exists(os.path.join(trial_path, "summary.csv")):
			raise ValueError(f"Trial path {trial_path} is already completed.")

		# Extract node lines from config.yaml
		yaml_path = os.path.join(trial_path, "config.yaml")
		node_lines = self._load_node_lines(yaml_path)

		node_line_names = list(node_lines.keys())
		nodes = list(node_lines.values())
		node_names = list(
			map(lambda node: list(map(lambda n: n.node_type, node)), nodes)
		)

		# If the First Node Line folder hasn't even been created, proceed to start_trial
		if not os.path.exists(os.path.join(trial_path, node_line_names[0])):
			self.start_trial(yaml_path)
			return None

		# Find conflict node line and node
		conflict_line_name, conflict_node_name = self.__find_conflict_point(
			trial_path, node_line_names, node_lines
		)
		node_dir = os.path.join(trial_path, conflict_line_name, conflict_node_name)
		if os.path.exists(node_dir):
			shutil.rmtree(node_dir)

		# Set remain_nodes and remain_lines
		remain_nodes, completed_node_names, remain_lines, remain_line_names = (
			self._set_remain_nodes_and_lines(
				node_line_names,
				nodes,
				node_names,
				conflict_node_name,
				conflict_line_name,
			)
		)
		# Set previous_result
		previous_result = self.__set_previous_result(
			node_line_names, node_names, trial_path, conflict_node_name
		)

		# Run Node
		if remain_nodes:
			conflict_line_dir = os.path.join(trial_path, conflict_line_name)
			summary_lst = []
			# Get already run node summary and append to summary_lst
			for completed_node_name in completed_node_names:
				summary_lst = self._append_node_summary(
					conflict_line_dir, completed_node_name, summary_lst
				)
			for node in remain_nodes:
				previous_result = node.run(previous_result, conflict_line_dir)
				summary_lst = self._append_node_summary(
					conflict_line_dir, node.node_type, summary_lst
				)
			pd.DataFrame(summary_lst).to_csv(
				os.path.join(conflict_line_dir, "summary.csv"), index=False
			)

		# Run node line
		trial_summary_df = pd.DataFrame(
			columns=[
				"node_line_name",
				"node_type",
				"best_module_filename",
				"best_module_name",
				"best_module_params",
				"best_execution_time",
			]
		)
		completed_line_names = node_line_names[
			: node_line_names.index(conflict_line_name)
		]
		# Get already run node line's summary and append to trial_summary_df
		if completed_line_names:
			for line_name in completed_line_names:
				node_line_dir = os.path.join(trial_path, line_name)
				trial_summary_df = self._append_node_line_summary(
					line_name, node_line_dir, trial_summary_df
				)
		if remain_lines:
			for node_line_name, node_line in zip(remain_line_names, remain_lines):
				node_line_dir = os.path.join(trial_path, node_line_name)
				if not os.path.exists(node_line_dir):
					os.makedirs(node_line_dir)
				logger.info(f"Running node line {node_line_name}...")
				previous_result = run_node_line(
					node_line, node_line_dir, previous_result
				)
				trial_summary_df = self._append_node_line_summary(
					node_line_name, node_line_dir, trial_summary_df
				)
		trial_summary_df.to_csv(os.path.join(trial_path, "summary.csv"), index=False)

		logger.info("Evaluation complete.")

	def __find_conflict_point(
		self,
		trial_path: str,
		node_line_names: List[str],
		node_lines: Dict[str, List[Node]],
	) -> tuple[str, str]:
		for node_line_name in node_line_names:
			node_line_dir = os.path.join(trial_path, node_line_name)
			if not os.path.exists(node_line_dir):
				return node_line_name, node_lines[node_line_name][0].node_type

			if not os.path.exists(os.path.join(node_line_dir, "summary.csv")):
				conflict_node_name = self._find_conflict_node_name(
					node_line_dir, node_lines[node_line_name]
				)
				return node_line_name, conflict_node_name

		raise ValueError(f"No error node line found in {trial_path}.")

	@staticmethod
	def _find_conflict_node_name(node_line_dir: str, node_line: List[Node]) -> str:
		for node in node_line:
			node_dir = os.path.join(node_line_dir, node.node_type)
			if not os.path.exists(node_dir) or not os.path.exists(
				os.path.join(node_dir, "summary.csv")
			):
				return node.node_type
		raise TypeError("No conflict node name found.")

	def __set_previous_result(
		self,
		node_line_names: List[str],
		node_names: List[List[str]],
		trial_path: str,
		conflict_node_name: str,
	):
		exploded_node_line, exploded_node = explode(node_line_names, node_names)
		conflict_node_index = exploded_node.index(conflict_node_name)
		# Set previous_result
		if conflict_node_index == 0:
			previous_result = self.qa_data
		else:
			previous_node_line = exploded_node_line[conflict_node_index - 1]
			previous_node = exploded_node[conflict_node_index - 1]

			previous_node_dir = os.path.join(
				trial_path, previous_node_line, previous_node
			)
			best_file_pattern = f"{previous_node_dir}/best_*.parquet"
			previous_result = pd.read_parquet(
				glob.glob(best_file_pattern)[0], engine="pyarrow"
			)
		return previous_result

	@staticmethod
	def _set_remain_nodes_and_lines(
		node_line_names: List[str],
		nodes: List[List[Node]],
		node_names: List[List[str]],
		conflict_node_name: str,
		conflict_node_line_name: str,
	):
		conflict_node_line_index = node_line_names.index(conflict_node_line_name)
		full_conflict_node_line_nodes = nodes[conflict_node_line_index]
		full_conflict_node_line_node_names = node_names[conflict_node_line_index]

		if conflict_node_name == full_conflict_node_line_node_names[0]:
			remain_nodes = None
			completed_node_names = None
			remain_node_lines = nodes[conflict_node_line_index:]
			remain_node_line_names = node_line_names[conflict_node_line_index:]
		else:
			conflict_node_index = full_conflict_node_line_node_names.index(
				conflict_node_name
			)
			remain_nodes = full_conflict_node_line_nodes[conflict_node_index:]
			completed_node_names = full_conflict_node_line_node_names[
				:conflict_node_index
			]
			if conflict_node_line_index + 1 >= len(node_line_names):
				remain_node_lines = None
				remain_node_line_names = None
			else:
				remain_node_lines = nodes[conflict_node_line_index + 1 :]
				remain_node_line_names = node_line_names[conflict_node_line_index + 1 :]
		return (
			remain_nodes,
			completed_node_names,
			remain_node_lines,
			remain_node_line_names,
		)

	@staticmethod
	def _append_node_line_summary(
		node_line_name: str, node_line_dir: str, trial_summary_df: pd.DataFrame
	):
		summary_df = load_summary_file(
			os.path.join(node_line_dir, "summary.csv"),
			dict_columns=["best_module_params"],
		)
		summary_df = summary_df.assign(node_line_name=node_line_name)
		summary_df = summary_df[list(trial_summary_df.columns)]
		if len(trial_summary_df) <= 0:
			trial_summary_df = summary_df
		else:
			trial_summary_df = pd.concat(
				[trial_summary_df, summary_df], ignore_index=True
			)
		return trial_summary_df

	@staticmethod
	def _append_node_summary(
		node_line_dir: str, node_name: str, summary_lst: List[Dict]
	):
		node_summary_df = load_summary_file(
			os.path.join(node_line_dir, node_name, "summary.csv")
		)
		best_node_row = node_summary_df.loc[node_summary_df["is_best"]]
		summary_lst.append(
			{
				"node_type": node_name,
				"best_module_filename": best_node_row["filename"].values[0],
				"best_module_name": best_node_row["module_name"].values[0],
				"best_module_params": best_node_row["module_params"].values[0],
				"best_execution_time": best_node_row["execution_time"].values[0],
			}
		)
		return summary_lst

	@staticmethod
	def _find_bm25_tokenizer(nodes: List[Node]):
		bm25_tokenizer_list = extract_values_from_nodes(nodes, "bm25_tokenizer")
		strategy_tokenizer_list = list(
			chain.from_iterable(
				extract_values_from_nodes_strategy(nodes, "bm25_tokenizer")
			)
		)
		return list(set(bm25_tokenizer_list + strategy_tokenizer_list))

	@staticmethod
	def _find_embedding_model(nodes: List[Node]):
		embedding_models_list = extract_values_from_nodes(nodes, "embedding_model")
		retrieval_module_dicts = extract_values_from_nodes_strategy(
			nodes, "retrieval_modules"
		)
		for retrieval_modules in retrieval_module_dicts:
			vectordb_modules = list(
				filter(lambda x: x["module_type"] == "vectordb", retrieval_modules)
			)
			embedding_models_list.extend(
				list(map(lambda x: x.get("embedding_model", None), vectordb_modules))
			)
		embedding_models_list = list(
			filter(lambda x: x is not None, embedding_models_list)
		)
		return list(set(embedding_models_list))

	async def __get_ingest_target_corpus(
		self, vectordb, full_ingest: bool
	) -> pd.DataFrame:
		if full_ingest is True:
			# get the target ingest corpus from the whole corpus
			target_corpus = await filter_exist_ids(vectordb, self.corpus_data)
		else:
			# get the target ingest corpus from the retrieval gt only
			target_corpus = await filter_exist_ids_from_retrieval_gt(
				vectordb, self.qa_data, self.corpus_data
			)
		return target_corpus



import logging
import os
from typing import List, Dict

import pandas as pd

from autorag.nodes.promptmaker.base import BasePromptMaker
from autorag.utils import result_to_dataframe, fetch_contents

logger = logging.getLogger("AutoRAG")


class WindowReplacement(BasePromptMaker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir, *args, **kwargs)
		# load corpus
		data_dir = os.path.join(project_dir, "data")
		self.corpus_data = pd.read_parquet(
			os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
		)

	@result_to_dataframe(["prompts"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		query, retrieved_contents, prompt = self.cast_to_run(
			previous_result, *args, **kwargs
		)
		retrieved_ids = previous_result["retrieved_ids"].tolist()
		# get metadata from corpus
		retrieved_metadata = fetch_contents(
			self.corpus_data, retrieved_ids, column_name="metadata"
		)
		return self._pure(prompt, query, retrieved_contents, retrieved_metadata)

	def _pure(
		self,
		prompt: str,
		queries: List[str],
		retrieved_contents: List[List[str]],
		retrieved_metadata: List[List[Dict]],
	) -> List[str]:
		"""
		Replace retrieved_contents with a window to create a Prompt
		(only available for corpus chunked with Sentence window method)
		You must type a prompt or prompt list at a config YAML file like this:

		.. Code:: yaml
		nodes:
		- node_type: prompt_maker
		  modules:
		  - module_type: window_replacement
		    prompt: [Answer this question: {query} \n\n {retrieved_contents},
		    Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

		:param prompt: A prompt string.
		:param queries: List of query strings.
		:param retrieved_contents: List of retrieved contents.
		:param retrieved_metadata: List of retrieved metadata.
		:return: Prompts that are made by window_replacement.
		"""

		def window_replacement_row(
			_prompt: str,
			_query: str,
			_retrieved_contents,
			_retrieved_metadata: List[Dict],
		) -> str:
			window_list = []
			for content, metadata in zip(_retrieved_contents, _retrieved_metadata):
				if "window" in metadata:
					window_list.append(metadata["window"])
				else:
					window_list.append(content)
					logger.info(
						"Only available for corpus chunked with Sentence window method."
						"window_replacement will not proceed."
					)
			contents_str = "\n\n".join(window_list)
			return _prompt.format(query=_query, retrieved_contents=contents_str)

		return list(
			map(
				lambda x: window_replacement_row(prompt, x[0], x[1], x[2]),
				zip(queries, retrieved_contents, retrieved_metadata),
			)
		)



import os.path
from itertools import chain
from typing import Tuple, List, Dict, Any, Optional

from llama_index.core import Document
from llama_index.core.node_parser.interface import NodeParser

from autorag.utils.util import process_batch, get_event_loop
from autorag.data.chunk.base import chunker_node, add_file_name
from autorag.data.utils.util import (
	add_essential_metadata_llama_text_node,
	get_start_end_idx,
)


@chunker_node
def llama_index_chunk(
	texts: List[str],
	chunker: NodeParser,
	file_name_language: Optional[str] = None,
	metadata_list: Optional[List[Dict[str, str]]] = None,
	batch: int = 8,
) -> Tuple[
	List[str], List[str], List[str], List[Tuple[int, int]], List[Dict[str, Any]]
]:
	"""
	Chunk texts from the parsed result to use llama index chunk method

	:param texts: The list of texts to chunk from the parsed result
	:param chunker: A llama index NodeParser(Chunker) instance.
	:param file_name_language: The language to use 'add_file_name' feature.
	    You need to set one of 'English' and 'Korean'
	    The 'add_file_name' feature is to add a file_name to chunked_contents.
	    This is used to prevent hallucination by retrieving contents from the wrong document.
	    Default form of 'English' is "file_name: {file_name}\n contents: {content}"
	:param metadata_list: The list of dict of metadata from the parsed result
	:param batch: The batch size for chunk texts. Default is 8
	:return: tuple of lists containing the chunked doc_id, contents, path, start_idx, end_idx and metadata
	"""
	tasks = [
		llama_index_chunk_pure(text, chunker, file_name_language, meta)
		for text, meta in zip(texts, metadata_list)
	]
	loop = get_event_loop()
	results = loop.run_until_complete(process_batch(tasks, batch))

	doc_id, contents, path, start_end_idx, metadata = (
		list(chain.from_iterable(item)) for item in zip(*results)
	)

	return list(doc_id), list(contents), list(path), list(start_end_idx), list(metadata)


async def llama_index_chunk_pure(
	text: str,
	chunker: NodeParser,
	file_name_language: Optional[str] = None,
	_metadata: Optional[Dict[str, str]] = None,
):
	# set document
	document = [Document(text=text, metadata=_metadata)]

	# chunk document
	chunk_results = await chunker.aget_nodes_from_documents(documents=document)

	# make doc_id
	doc_id = list(map(lambda node: node.node_id, chunk_results))

	# make path
	path_lst = list(map(lambda x: x.metadata.get("path", ""), chunk_results))

	# make contents and start_end_idx
	if file_name_language:
		chunked_file_names = list(map(lambda x: os.path.basename(x), path_lst))
		chunked_texts = list(map(lambda x: x.text, chunk_results))
		start_end_idx = list(
			map(
				lambda x: get_start_end_idx(text, x),
				chunked_texts,
			)
		)
		contents = add_file_name(file_name_language, chunked_file_names, chunked_texts)
	else:
		contents = list(map(lambda x: x.text, chunk_results))
		start_end_idx = list(map(lambda x: get_start_end_idx(text, x), contents))

	metadata = list(
		map(
			lambda node: add_essential_metadata_llama_text_node(
				node.metadata, node.relationships
			),
			chunk_results,
		)
	)

	return doc_id, contents, path_lst, start_end_idx, metadata



from datetime import datetime

import numpy as np
import pandas as pd

from autorag.utils.util import preprocess_text


def validate_qa_dataset(df: pd.DataFrame):
	columns = ["qid", "query", "retrieval_gt", "generation_gt"]
	assert set(columns).issubset(
		df.columns
	), f"df must have columns {columns}, but got {df.columns}"


def validate_corpus_dataset(df: pd.DataFrame):
	columns = ["doc_id", "contents", "metadata"]
	assert set(columns).issubset(
		df.columns
	), f"df must have columns {columns}, but got {df.columns}"


def cast_qa_dataset(df: pd.DataFrame):
	def cast_retrieval_gt(gt):
		if isinstance(gt, str):
			return [[gt]]
		elif isinstance(gt, list):
			if isinstance(gt[0], str):
				return [gt]
			elif isinstance(gt[0], list):
				return gt
			elif isinstance(gt[0], np.ndarray):
				return cast_retrieval_gt(list(map(lambda x: x.tolist(), gt)))
			else:
				raise ValueError(
					f"retrieval_gt must be str or list, but got {type(gt[0])}"
				)
		elif isinstance(gt, np.ndarray):
			return cast_retrieval_gt(gt.tolist())
		else:
			raise ValueError(f"retrieval_gt must be str or list, but got {type(gt)}")

	def cast_generation_gt(gt):
		if isinstance(gt, str):
			return [gt]
		elif isinstance(gt, list):
			return gt
		elif isinstance(gt, np.ndarray):
			return cast_generation_gt(gt.tolist())
		else:
			raise ValueError(f"generation_gt must be str or list, but got {type(gt)}")

	df = df.reset_index(drop=True)
	validate_qa_dataset(df)
	assert df["qid"].apply(lambda x: isinstance(x, str)).sum() == len(
		df
	), "qid must be string type."
	assert df["query"].apply(lambda x: isinstance(x, str)).sum() == len(
		df
	), "query must be string type."
	df["retrieval_gt"] = df["retrieval_gt"].apply(cast_retrieval_gt)
	df["generation_gt"] = df["generation_gt"].apply(cast_generation_gt)
	df["query"] = df["query"].apply(preprocess_text)
	df["generation_gt"] = df["generation_gt"].apply(
		lambda x: list(map(preprocess_text, x))
	)
	return df


def cast_corpus_dataset(df: pd.DataFrame):
	df = df.reset_index(drop=True)
	validate_corpus_dataset(df)

	# drop rows that have empty contents
	df = df[~df["contents"].apply(lambda x: x is None or x.isspace())]

	def make_datetime_metadata(x):
		if x is None or x == {}:
			return {"last_modified_datetime": datetime.now()}
		elif x.get("last_modified_datetime") is None:
			return {**x, "last_modified_datetime": datetime.now()}
		else:
			return x

	df["metadata"] = df["metadata"].apply(make_datetime_metadata)

	# check every metadata have a datetime key
	assert sum(
		df["metadata"].apply(lambda x: x.get("last_modified_datetime") is not None)
	) == len(df), "Every metadata must have a datetime key."

	def make_prev_next_id_metadata(x, id_type: str):
		if x is None or x == {}:
			return {id_type: None}
		elif x.get(id_type) is None:
			return {**x, id_type: None}
		else:
			return x

	df["metadata"] = df["metadata"].apply(
		lambda x: make_prev_next_id_metadata(x, "prev_id")
	)
	df["metadata"] = df["metadata"].apply(
		lambda x: make_prev_next_id_metadata(x, "next_id")
	)

	df["contents"] = df["contents"].apply(preprocess_text)

	def normalize_unicode_metadata(metadata: dict):
		result = {}
		for key, value in metadata.items():
			if isinstance(value, str):
				result[key] = preprocess_text(value)
			else:
				result[key] = value
		return result

	df["metadata"] = df["metadata"].apply(normalize_unicode_metadata)

	# check every metadata have a prev_id, next_id key
	assert all(
		"prev_id" in metadata for metadata in df["metadata"]
	), "Every metadata must have a prev_id key."
	assert all(
		"next_id" in metadata for metadata in df["metadata"]
	), "Every metadata must have a next_id key."

	return df


def validate_qa_from_corpus_dataset(qa_df: pd.DataFrame, corpus_df: pd.DataFrame):
	qa_ids = []
	for retrieval_gt in qa_df["retrieval_gt"].tolist():
		if isinstance(retrieval_gt, list) and (
			retrieval_gt[0] != [] or any(bool(g) is True for g in retrieval_gt)
		):
			for gt in retrieval_gt:
				qa_ids.extend(gt)
		elif isinstance(retrieval_gt, np.ndarray) and retrieval_gt[0].size > 0:
			for gt in retrieval_gt:
				qa_ids.extend(gt)

	no_exist_ids = list(
		filter(lambda qa_id: corpus_df[corpus_df["doc_id"] == qa_id].empty, qa_ids)
	)

	assert (
		len(no_exist_ids) == 0
	), f"{len(no_exist_ids)} doc_ids in retrieval_gt do not exist in corpus_df."



from abc import abstractmethod
from typing import List, Tuple, Union

from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.utils.util import openai_truncate_by_token
from autorag.embedding.base import EmbeddingModel


class BaseVectorStore:
	support_similarity_metrics = ["l2", "ip", "cosine"]

	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		similarity_metric: str = "cosine",
		embedding_batch: int = 100,
	):
		self.embedding = EmbeddingModel.load(embedding_model)()
		self.embedding_batch = embedding_batch
		self.embedding.embed_batch_size = embedding_batch
		assert (
			similarity_metric in self.support_similarity_metrics
		), f"search method {similarity_metric} is not supported"
		self.similarity_metric = similarity_metric

	@abstractmethod
	async def add(
		self,
		ids: List[str],
		texts: List[str],
	):
		pass

	@abstractmethod
	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		"""
		Add the embeddings to the Vector DB.
		"""
		pass

	@abstractmethod
	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		pass

	@abstractmethod
	async def fetch(self, ids: List[str]) -> List[List[float]]:
		"""
		Fetch the embeddings of the ids.
		"""
		pass

	@abstractmethod
	async def is_exist(self, ids: List[str]) -> List[bool]:
		"""
		Check if the ids exist in the Vector DB.
		"""
		pass

	@abstractmethod
	async def delete(self, ids: List[str]):
		pass

	def truncated_inputs(self, inputs: List[str]) -> List[str]:
		if isinstance(self.embedding, OpenAIEmbedding):
			openai_embedding_limit = 8000
			results = openai_truncate_by_token(
				inputs, openai_embedding_limit, self.embedding.model_name
			)
			return results
		return inputs



import logging
from typing import List, Tuple
import time

import pandas as pd
import requests
from asyncio import to_thread

from autorag.nodes.generator.base import BaseGenerator
from autorag.utils.util import get_event_loop, process_batch, result_to_dataframe

logger = logging.getLogger("AutoRAG")

DEFAULT_MAX_TOKENS = 4096  # Default token limit


class VllmAPI(BaseGenerator):
	def __init__(
		self,
		project_dir,
		llm: str,
		uri: str,
		max_tokens: int = None,
		batch: int = 16,
		*args,
		**kwargs,
	):
		"""
		VLLM API Wrapper for OpenAI-compatible chat/completions format.

		:param project_dir: Project directory.
		:param llm: Model name (e.g., LLaMA model).
		:param uri: VLLM API server URI.
		:param max_tokens: Maximum token limit.
		    Default is 4096.
		:param batch: Request batch size.
		    Default is 16.
		"""
		super().__init__(project_dir, llm, *args, **kwargs)
		assert batch > 0, "Batch size must be greater than 0."
		self.uri = uri.rstrip("/")  # Set API URI
		self.batch = batch
		# Use the provided max_tokens if available, otherwise use the default
		self.max_token_size = max_tokens if max_tokens else DEFAULT_MAX_TOKENS
		self.max_model_len = self.get_max_model_length()
		logger.info(f"{llm} max model length: {self.max_model_len}")

	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		prompts = self.cast_to_run(previous_result)
		return self._pure(prompts, **kwargs)

	def _pure(
		self, prompts: List[str], truncate: bool = True, **kwargs
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		Method to call the VLLM API to generate text.

		:param prompts: List of input prompts.
		:param truncate: Whether to truncate input prompts to fit within the token limit.
		:param kwargs: Additional options (e.g., temperature, top_p).
		:return: Generated text, token lists, and log probability lists.
		"""
		if kwargs.get("logprobs") is not None:
			kwargs.pop("logprobs")
			logger.warning(
				"parameter logprob does not effective. It always set to True."
			)
		if kwargs.get("n") is not None:
			kwargs.pop("n")
			logger.warning("parameter n does not effective. It always set to 1.")

		if truncate:
			prompts = list(map(lambda p: self.truncate_by_token(p), prompts))
		loop = get_event_loop()
		tasks = [to_thread(self.get_result, prompt, **kwargs) for prompt in prompts]
		results = loop.run_until_complete(process_batch(tasks, self.batch))

		answer_result = list(map(lambda x: x[0], results))
		token_result = list(map(lambda x: x[1], results))
		logprob_result = list(map(lambda x: x[2], results))
		return answer_result, token_result, logprob_result

	def truncate_by_token(self, prompt: str) -> str:
		"""
		Function to truncate prompts to fit within the maximum token limit.
		"""
		tokens = self.encoding_for_model(prompt)["tokens"]  # Simple tokenization
		return self.decoding_for_model(tokens[: self.max_model_len])["prompt"]

	def call_vllm_api(self, prompt: str, **kwargs) -> dict:
		"""
		Calls the VLLM API to get chat/completions responses.

		:param prompt: Input prompt.
		:param kwargs: Additional API options (e.g., temperature, max_tokens).
		:return: API response.
		"""
		payload = {
			"model": self.llm,
			"messages": [{"role": "user", "content": prompt}],
			"temperature": kwargs.get("temperature", 0.4),
			"max_tokens": min(
				kwargs.get("max_tokens", self.max_token_size), self.max_token_size
			),
			"logprobs": True,
			"n": 1,
		}
		start_time = time.time()  # Record request start time
		response = requests.post(f"{self.uri}/v1/chat/completions", json=payload)
		end_time = time.time()  # Record request end time

		response.raise_for_status()
		elapsed_time = end_time - start_time  # Calculate elapsed time
		logger.info(
			f"Request chat completions to vllm server completed in {elapsed_time:.2f} seconds"
		)
		return response.json()

	# Additional method: abstract method implementation
	async def astream(self, prompt: str, **kwargs):
		"""
		Asynchronous streaming method not implemented.
		"""
		raise NotImplementedError("astream method is not implemented for VLLM API yet.")

	def stream(self, prompt: str, **kwargs):
		"""
		Synchronous streaming method not implemented.
		"""
		raise NotImplementedError("stream method is not implemented for VLLM API yet.")

	def get_result(self, prompt: str, **kwargs):
		response = self.call_vllm_api(prompt, **kwargs)
		choice = response["choices"][0]
		answer = choice["message"]["content"]

		# Handle cases where logprobs is None
		if choice.get("logprobs") and "content" in choice["logprobs"]:
			logprobs = list(map(lambda x: x["logprob"], choice["logprobs"]["content"]))
			tokens = list(
				map(
					lambda x: self.encoding_for_model(x["token"])["tokens"],
					choice["logprobs"]["content"],
				)
			)
		else:
			logprobs = []
			tokens = []

		return answer, tokens, logprobs

	def encoding_for_model(self, answer_piece: str):
		payload = {
			"model": self.llm,
			"prompt": answer_piece,
			"add_special_tokens": True,
		}
		response = requests.post(f"{self.uri}/tokenize", json=payload)
		response.raise_for_status()
		return response.json()

	def decoding_for_model(self, tokens: list[int]):
		payload = {
			"model": self.llm,
			"tokens": tokens,
		}
		response = requests.post(f"{self.uri}/detokenize", json=payload)
		response.raise_for_status()
		return response.json()

	def get_max_model_length(self):
		response = requests.get(f"{self.uri}/v1/models")
		response.raise_for_status()
		json_data = response.json()
		return json_data["data"][0]["max_model_len"]



import os.path
import random
from typing import Optional, List, Dict, Any

import pandas as pd
from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.llms import LLM

from autorag.utils.util import process_batch, get_event_loop

package_dir = os.path.dirname(os.path.realpath(__file__))


def generate_qa_llama_index(
	llm: LLM,
	contents: List[str],
	prompt: Optional[str] = None,
	question_num_per_content: int = 1,
	max_retries: int = 3,
	batch: int = 4,
) -> List[List[Dict]]:
	"""
	Generate a qa set from the list of contents.
	It uses a single prompt for all contents.
	If you want to use more than one prompt for generating qa,
	you can consider using generate_qa_llama_index_by_ratio.

	:param llm: Llama index model
	:param contents: List of content strings.
	:param prompt: The prompt to use for the qa generation.
	    The prompt must include the following placeholders:
	    - {{text}}: The content string
	    - {{num_questions}}: The number of questions to generate
	    As default, the prompt is set to the default prompt for the question type.
	:param question_num_per_content: Number of questions to generate for each content.
	    Default is 1.
	:param max_retries: The maximum number of retries when generated question number is not equal to the target number.
	    Default is 3.
	:param batch: The batch size to process asynchronously.
	    Default is 4.
	:return: 2-d list of dictionaries containing the query and generation_gt.
	"""
	# load default prompt
	if prompt is None:
		prompt = open(
			os.path.join(package_dir, "llama_index_default_prompt.txt"), "r"
		).read()

	tasks = [
		async_qa_gen_llama_index(
			content, llm, prompt, question_num_per_content, max_retries
		)
		for content in contents
	]
	loops = get_event_loop()
	results = loops.run_until_complete(process_batch(tasks, batch))
	return results


def generate_answers(
	llm: LLM,
	contents: List[str],
	queries: List[str],
	batch: int = 4,
) -> List[List[Dict]]:
	"""
	Generate qa sets from the list of contents using existing queries.

	:param llm: Llama index model
	:param contents: List of content strings.
	:param queries: List of existing queries.
	:param batch: The batch size to process asynchronously.
	:return: 2-d list of dictionaries containing the query and generation_gt.
	"""

	tasks = [
		generate_basic_answer(llm, content, query)
		for content, query in zip(contents, queries)
	]
	loops = get_event_loop()
	results = loops.run_until_complete(process_batch(tasks, batch))
	return results


def generate_qa_llama_index_by_ratio(
	llm: LLM,
	contents: List[str],
	prompts_ratio: Dict,
	question_num_per_content: int = 1,
	max_retries: int = 3,
	random_state: int = 42,
	batch: int = 4,
) -> List[List[Dict]]:
	"""
	Generate a qa set from the list of contents.
	You can set the ratio of prompts that you want to use for generating qa.
	It distributes the number of questions to generate for each content by the ratio randomly.

	:param llm: Llama index model
	:param contents: List of content strings.
	:param prompts_ratio: Dictionary of prompt paths and their ratios.
	    Example: {"prompt/prompt1.txt": 0.5, "prompt/prompt2.txt": 0.5}
	    The value sum doesn't have to be 1.
	    The path must be the absolute path, and the file must exist.
	    Plus, it has to be a text file which contains proper prompt.
	    Each prompt must contain the following placeholders:
	    - {{text}}: The content string
	    - {{num_questions}}: The number of questions to generate
	:param question_num_per_content: Number of questions to generate for each content.
	    Default is 1.
	:param max_retries: The maximum number of retries when generated question number is not equal to the target number.
	    Default is 3.
	:param random_state: Random seed
	    Default is 42.
	:param batch: The batch size to process asynchronously.
	    Default is 4.
	:return: 2-d list of dictionaries containing the query and generation_gt.
	"""
	prompts = list(map(lambda path: open(path, "r").read(), prompts_ratio.keys()))
	assert all([validate_llama_index_prompt(prompt) for prompt in prompts])

	content_indices = list(range(len(contents)))
	random.seed(random_state)
	random.shuffle(content_indices)

	slice_content_indices: List[List[str]] = distribute_list_by_ratio(
		content_indices, list(prompts_ratio.values())
	)
	temp_df = pd.DataFrame({"idx": slice_content_indices, "prompt": prompts})
	temp_df = temp_df.explode("idx", ignore_index=True)
	temp_df = temp_df.sort_values(by="idx", ascending=True)

	final_df = pd.DataFrame({"content": contents, "prompt": temp_df["prompt"].tolist()})

	tasks = [
		async_qa_gen_llama_index(
			content, llm, prompt, question_num_per_content, max_retries
		)
		for content, prompt in zip(
			final_df["content"].tolist(), final_df["prompt"].tolist()
		)
	]

	loops = get_event_loop()
	results = loops.run_until_complete(process_batch(tasks, batch))

	return results


async def async_qa_gen_llama_index(
	content: str,
	llm: LLM,
	prompt: str,
	question_num: int = 1,
	max_retries: int = 3,
):
	"""
	Generate a qa set by using the given content and the llama index model.
	You must select the question type.

	:param content: Content string
	:param llm: Llama index model
	:param prompt: The prompt to use for the qa generation.
	    The prompt must include the following placeholders:
	    - {{text}}: The content string
	    - {{num_questions}}: The number of questions to generate
	:param question_num: The number of questions to generate
	:param max_retries: Maximum number of retries when generated question number is not equal to the target number
	:return: List of dictionaries containing the query and generation_gt
	"""
	validate_llama_index_prompt(prompt)

	async def generate(content: str, llm: LLM):
		for _ in range(max_retries):
			output = await llm.acomplete(
				prompt.replace("{{text}}", content).replace(
					"{{num_questions}}", str(question_num)
				)
			)
			result = parse_output(output.text)
			if len(result) == question_num:
				return result
		raise InterruptedError(
			f"Failed to generate output of length {question_num} after {max_retries} retries."
		)

	return await generate(content, llm)


async def generate_basic_answer(llm: LLM, passage_str: str, query: str) -> str:
	basic_answer_system_prompt = """You are an AI assistant to answer the given question in the provide evidence text.
    You can find the evidence from the given text about question, and you have to write a proper answer to the given question.
    You have to preserve the question's language at the answer.
    For example, if the input question is Korean, the output answer must be in Korean.
    """
	user_prompt = f"Text:\n<|text_start|>\n{passage_str}\n<|text_end|>\n\nQuestion:\n{query}\n\nAnswer:"

	response = await llm.achat(
		messages=[
			ChatMessage(role=MessageRole.SYSTEM, content=basic_answer_system_prompt),
			ChatMessage(role=MessageRole.USER, content=user_prompt),
		],
		temperature=1.0,
	)
	return response.message.content


def validate_llama_index_prompt(prompt: str) -> bool:
	"""
	Validate the prompt for the llama index model.
	The prompt must include the following placeholders:
	- {{text}}: The content string
	- {{num_questions}}: The number of questions to generate
	"""
	if "{{text}}" not in prompt:
		raise ValueError("The prompt must include the placeholder {{text}}.")
	if "{{num_questions}}" not in prompt:
		raise ValueError("The prompt must include the placeholder {{num_questions}}.")
	return True


def parse_output(result: str) -> List[Dict]:
	result = result.strip()
	result = result.split("[Q]:")
	final_result = list()
	for res in result:
		res = res.strip()
		if res and "\n[A]:" in res:
			qa = res.split("\n[A]:")
			final_result.append(
				{"query": qa[0].strip(), "generation_gt": qa[1].strip()}
			)
	return final_result


def distribute_list_by_ratio(input_list, ratio) -> List[List[Any]]:
	total_ratio = sum(ratio)
	total_length = len(input_list)

	# Calculate the length of each slice
	slice_lengths = [int((r / total_ratio) * total_length) for r in ratio]

	# Adjust the last slice in case of rounding issues
	slice_lengths[-1] = total_length - sum(slice_lengths[:-1])

	slices = []
	start = 0
	for length in slice_lengths:
		end = start + length
		slices.append(input_list[start:end])
		start = end

	return slices



from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	flatten_apply,
	make_batch,
	select_top_k,
	sort_by_scores,
	pop_params,
	result_to_dataframe,
	empty_cuda_cache,
)


class SentenceTransformerReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model_name: str = "cross-encoder/ms-marco-MiniLM-L-2-v2",
		*args,
		**kwargs,
	):
		"""
		Initialize the Sentence Transformer reranker node.

		:param project_dir: The project directory
		:param model_name: The name of the Sentence Transformer model to use for reranking
		Default is "cross-encoder/ms-marco-MiniLM-L-2-v2"
		:param kwargs: The CrossEncoder parameters
		"""
		super().__init__(project_dir, *args, **kwargs)
		try:
			import torch
			from sentence_transformers import CrossEncoder
		except ImportError:
			raise ImportError(
				"You have to install AutoRAG[gpu] to use SentenceTransformerReranker"
			)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		model_params = pop_params(CrossEncoder.__init__, kwargs)
		self.model = CrossEncoder(model_name, device=self.device, **model_params)

	def __del__(self):
		del self.model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		"""
		Rerank a list of contents based on their relevance to a query using a Sentence Transformer model.

		:param previous_result: The previous result
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
		:return: pd DataFrame containing the reranked contents, ids, and scores
		"""
		queries, contents_list, scores_list, ids_list = self.cast_to_run(
			previous_result
		)
		top_k = kwargs.get("top_k", 1)
		batch = kwargs.get("batch", 64)
		return self._pure(queries, contents_list, ids_list, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using a Sentence Transformer model.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch

		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]
		rerank_scores = flatten_apply(
			sentence_transformer_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def sentence_transformer_run_model(input_texts, model, batch_size: int):
	try:
		import torch
	except ImportError:
		raise ImportError(
			"You have to install AutoRAG[gpu] to use SentenceTransformerReranker"
		)
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in batch_input_texts:
		with torch.no_grad():
			pred_scores = model.predict(batch_texts)
		results.extend(pred_scores.tolist())
	return results



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import copy

from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.models.t5.modeling_t5 import T5Config, T5PreTrainedModel, T5Stack
from transformers.utils.model_parallel_utils import assert_device_map, get_device_map

from autorag.utils.util import empty_cuda_cache


class EncT5ForSequenceClassification(T5PreTrainedModel):
	_keys_to_ignore_on_load_missing = [
		r"encoder\.embed_tokens\.weight",
	]

	def __init__(self, config: T5Config, dropout=0.1):
		super().__init__(config)
		try:
			from torch import nn
		except ImportError:
			raise ImportError("Please install PyTorch to use TART reranker.")
		self.num_labels = config.num_labels
		self.config = config

		self.shared = nn.Embedding(config.vocab_size, config.d_model)

		encoder_config = copy.deepcopy(config)
		encoder_config.use_cache = False
		encoder_config.is_encoder_decoder = False
		self.encoder = T5Stack(encoder_config, self.shared)

		self.dropout = nn.Dropout(dropout)
		self.classifier = nn.Linear(config.hidden_size, config.num_labels)

		# Initialize weights and apply final processing
		self.post_init()

		# Model parallel
		self.model_parallel = False
		self.device_map = None

	def parallelize(self, device_map=None):
		try:
			import torch
		except ImportError:
			raise ImportError("Please install PyTorch to use TART reranker.")
		self.device_map = (
			get_device_map(len(self.encoder.block), range(torch.cuda.device_count()))
			if device_map is None
			else device_map
		)
		assert_device_map(self.device_map, len(self.encoder.block))
		self.encoder.parallelize(self.device_map)
		self.classifier = self.classifier.to(self.encoder.first_device)
		self.model_parallel = True

	def deparallelize(self):
		self.encoder.deparallelize()
		self.encoder = self.encoder.to("cpu")
		self.model_parallel = False
		self.device_map = None
		empty_cuda_cache()

	def get_input_embeddings(self):
		return self.shared

	def set_input_embeddings(self, new_embeddings):
		self.shared = new_embeddings
		self.encoder.set_input_embeddings(new_embeddings)

	def get_encoder(self):
		return self.encoder

	def _prune_heads(self, heads_to_prune):
		"""
		Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
		class PreTrainedModel
		"""
		for layer, heads in heads_to_prune.items():
			self.encoder.layer[layer].attention.prune_heads(heads)

	def forward(
		self,
		input_ids=None,
		attention_mask=None,
		head_mask=None,
		inputs_embeds=None,
		labels=None,
		output_attentions=None,
		output_hidden_states=None,
		return_dict=None,
	):
		try:
			import torch
			from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
		except ImportError:
			raise ImportError("Please install PyTorch to use TART reranker.")
		return_dict = (
			return_dict if return_dict is not None else self.config.use_return_dict
		)

		outputs = self.encoder(
			input_ids=input_ids,
			attention_mask=attention_mask,
			inputs_embeds=inputs_embeds,
			head_mask=head_mask,
			output_attentions=output_attentions,
			output_hidden_states=output_hidden_states,
			return_dict=return_dict,
		)

		hidden_states = outputs[0]
		pooled_output = hidden_states[:, 0, :]  # Take bos token (equiv. to <s>)

		pooled_output = self.dropout(pooled_output)
		logits = self.classifier(pooled_output)

		loss = None
		if labels is not None:
			if self.config.problem_type is None:
				if self.num_labels == 1:
					self.config.problem_type = "regression"
				elif self.num_labels > 1 and (
					labels.dtype == torch.long or labels.dtype == torch.int
				):
					self.config.problem_type = "single_label_classification"
				else:
					self.config.problem_type = "multi_label_classification"

			if self.config.problem_type == "regression":
				loss_fct = MSELoss()
				if self.num_labels == 1:
					loss = loss_fct(logits.squeeze(), labels.squeeze())
				else:
					loss = loss_fct(logits, labels)
			elif self.config.problem_type == "single_label_classification":
				loss_fct = CrossEntropyLoss()
				loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
			elif self.config.problem_type == "multi_label_classification":
				loss_fct = BCEWithLogitsLoss()
				loss = loss_fct(logits, labels)
		if not return_dict:
			output = (logits,) + outputs[1:]
			return ((loss,) + output) if loss is not None else output

		return SequenceClassifierOutput(
			loss=loss,
			logits=logits,
			hidden_states=outputs.hidden_states,
			attentions=outputs.attentions,
		)



from typing import List, Tuple

import numpy as np
import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	empty_cuda_cache,
)


class KoReranker(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir)
		try:
			import torch
			from transformers import AutoModelForSequenceClassification, AutoTokenizer
		except ImportError:
			raise ImportError("For using KoReranker, please install torch first.")

		model_path = "Dongjin-kr/ko-reranker"
		self.tokenizer = AutoTokenizer.from_pretrained(model_path)
		self.model = AutoModelForSequenceClassification.from_pretrained(model_path)
		self.model.eval()
		# Determine the device to run the model on (GPU if available, otherwise CPU)
		self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
		self.model.to(self.device)

	def __del__(self):
		del self.model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using ko-reranker.
		ko-reranker is a reranker based on korean (https://huggingface.co/Dongjin-kr/ko-reranker).

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
		    Default is 64.
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]
		scores_nps = flatten_apply(
			koreranker_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
			device=self.device,
		)

		rerank_scores = list(
			map(
				lambda scores: exp_normalize(np.array(scores)).astype(float), scores_nps
			)
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def koreranker_run_model(input_texts, model, tokenizer, device, batch_size: int):
	try:
		import torch
	except ImportError:
		raise ImportError("For using KoReranker, please install torch first.")
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in batch_input_texts:
		inputs = tokenizer(
			batch_texts,
			padding=True,
			truncation=True,
			return_tensors="pt",
			max_length=512,
		)
		inputs = inputs.to(device)
		with torch.no_grad():
			scores = (
				model(**inputs, return_dict=True)
				.logits.view(
					-1,
				)
				.float()
			)
			scores_np = scores.cpu().numpy()
			results.extend(scores_np)
	return results


def exp_normalize(x):
	b = x.max()
	y = np.exp(x - b)
	return y / y.sum()



import logging

from datetime import timedelta

from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.options import ClusterOptions

from typing import List, Tuple, Optional, Union

from autorag.utils.util import make_batch
from autorag.vectordb import BaseVectorStore

logger = logging.getLogger("AutoRAG")


class Couchbase(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		bucket_name: str,
		scope_name: str,
		collection_name: str,
		index_name: str,
		embedding_batch: int = 100,
		connection_string: str = "",
		username: str = "",
		password: str = "",
		ingest_batch: int = 100,
		text_key: Optional[str] = "text",
		embedding_key: Optional[str] = "embedding",
		scoped_index: bool = True,
	):
		super().__init__(
			embedding_model=embedding_model,
			similarity_metric="ip",
			embedding_batch=embedding_batch,
		)

		self.index_name = index_name
		self.bucket_name = bucket_name
		self.scope_name = scope_name
		self.collection_name = collection_name
		self.scoped_index = scoped_index
		self.text_key = text_key
		self.embedding_key = embedding_key
		self.ingest_batch = ingest_batch

		auth = PasswordAuthenticator(username, password)
		self.cluster = Cluster(connection_string, ClusterOptions(auth))

		# Wait until the cluster is ready for use.
		self.cluster.wait_until_ready(timedelta(seconds=5))

		# Check if the bucket exists
		if not self._check_bucket_exists():
			raise ValueError(
				f"Bucket {self.bucket_name} does not exist. "
				" Please create the bucket before searching."
			)

		try:
			self.bucket = self.cluster.bucket(self.bucket_name)
			self.scope = self.bucket.scope(self.scope_name)
			self.collection = self.scope.collection(self.collection_name)
		except Exception as e:
			raise ValueError(
				"Error connecting to couchbase. "
				"Please check the connection and credentials."
			) from e

		# Check if the index exists. Throws ValueError if it doesn't
		try:
			self._check_index_exists()
		except Exception:
			raise

		# Reinitialize to ensure a consistent state
		self.bucket = self.cluster.bucket(self.bucket_name)
		self.scope = self.bucket.scope(self.scope_name)
		self.collection = self.scope.collection(self.collection_name)

	async def add(self, ids: List[str], texts: List[str]):
		from couchbase.exceptions import DocumentExistsException

		texts = self.truncated_inputs(texts)
		text_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(texts)

		documents_to_insert = []
		for _id, text, embedding in zip(ids, texts, text_embeddings):
			doc = {
				self.text_key: text,
				self.embedding_key: embedding,
			}
			documents_to_insert.append({_id: doc})

		batch_documents_to_insert = make_batch(documents_to_insert, self.ingest_batch)

		for batch in batch_documents_to_insert:
			insert_batch = {}
			for doc in batch:
				insert_batch.update(doc)
			try:
				self.collection.upsert_multi(insert_batch)
			except DocumentExistsException as e:
				logger.debug(f"Document already exists: {e}")

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		from couchbase.exceptions import DocumentExistsException

		documents_to_insert = []
		for _id, embedding in zip(ids, embeddings):
			doc = {
				self.embedding_key: embedding,
			}
			documents_to_insert.append({_id: doc})

		batch_documents_to_insert = make_batch(documents_to_insert, self.ingest_batch)

		for batch in batch_documents_to_insert:
			insert_batch = {}
			for doc in batch:
				insert_batch.update(doc)
			try:
				self.collection.upsert_multi(insert_batch)
			except DocumentExistsException as e:
				logger.debug(f"Document already exists: {e}")

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		# Fetch vectors by IDs
		fetched_result = self.collection.get_multi(ids)
		fetched_vectors = {
			k: v.value[f"{self.embedding_key}"]
			for k, v in fetched_result.results.items()
		}
		return list(map(lambda x: fetched_vectors[x], ids))

	async def is_exist(self, ids: List[str]) -> List[bool]:
		existed_result = self.collection.exists_multi(ids)
		existed_ids = {k: v.exists for k, v in existed_result.results.items()}
		return list(map(lambda x: existed_ids[x], ids))

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		import couchbase.search as search
		from couchbase.options import SearchOptions
		from couchbase.vector_search import VectorQuery, VectorSearch

		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)

		ids, scores = [], []
		for query_embedding in query_embeddings:
			# Create Search Request
			search_req = search.SearchRequest.create(
				VectorSearch.from_vector_query(
					VectorQuery(
						self.embedding_key,
						query_embedding,
						top_k,
					)
				)
			)

			# Search
			if self.scoped_index:
				search_iter = self.scope.search(
					self.index_name,
					search_req,
					SearchOptions(limit=top_k),
				)

			else:
				search_iter = self.cluster.search(
					self.index_name,
					search_req,
					SearchOptions(limit=top_k),
				)

			# Parse the search results
			# search_iter.rows() can only be iterated once.
			id_list, score_list = [], []
			for result in search_iter.rows():
				id_list.append(result.id)
				score_list.append(result.score)

			ids.append(id_list)
			scores.append(score_list)

		return ids, scores

	async def delete(self, ids: List[str]):
		self.collection.remove_multi(ids)

	def _check_bucket_exists(self) -> bool:
		"""Check if the bucket exists in the linked Couchbase cluster.

		Returns:
		    True if the bucket exists
		"""
		bucket_manager = self.cluster.buckets()
		try:
			bucket_manager.get_bucket(self.bucket_name)
			return True
		except Exception as e:
			logger.debug("Error checking if bucket exists:", e)
			return False

	def _check_index_exists(self) -> bool:
		"""Check if the Search index exists in the linked Couchbase cluster
		Returns:
		    bool: True if the index exists, False otherwise.
		    Raises a ValueError if the index does not exist.
		"""
		if self.scoped_index:
			all_indexes = [
				index.name for index in self.scope.search_indexes().get_all_indexes()
			]
			if self.index_name not in all_indexes:
				raise ValueError(
					f"Index {self.index_name} does not exist. "
					" Please create the index before searching."
				)
		else:
			all_indexes = [
				index.name for index in self.cluster.search_indexes().get_all_indexes()
			]
			if self.index_name not in all_indexes:
				raise ValueError(
					f"Index {self.index_name} does not exist. "
					" Please create the index before searching."
				)

		return True



import logging
from typing import List, Callable

from langchain_community.document_loaders import (
	PDFMinerLoader,
	PDFPlumberLoader,
	PyPDFium2Loader,
	PyPDFLoader,
	PyMuPDFLoader,
	UnstructuredPDFLoader,
	CSVLoader,
	JSONLoader,
	UnstructuredMarkdownLoader,
	BSHTMLLoader,
	UnstructuredXMLLoader,
	DirectoryLoader,
)
from langchain_unstructured import UnstructuredLoader
from langchain_upstage import UpstageLayoutAnalysisLoader

from llama_index.core.node_parser import (
	TokenTextSplitter,
	SentenceSplitter,
	SentenceWindowNodeParser,
	SemanticSplitterNodeParser,
	SemanticDoubleMergingSplitterNodeParser,
	SimpleFileNodeParser,
)
from langchain.text_splitter import (
	RecursiveCharacterTextSplitter,
	CharacterTextSplitter,
	KonlpyTextSplitter,
	SentenceTransformersTokenTextSplitter,
)

from autorag import LazyInit

logger = logging.getLogger("AutoRAG")

parse_modules = {
	# PDF
	"pdfminer": PDFMinerLoader,
	"pdfplumber": PDFPlumberLoader,
	"pypdfium2": PyPDFium2Loader,
	"pypdf": PyPDFLoader,
	"pymupdf": PyMuPDFLoader,
	"unstructuredpdf": UnstructuredPDFLoader,
	# Common File Types
	# 1. CSV
	"csv": CSVLoader,
	# 2. JSON
	"json": JSONLoader,
	# 3. Markdown
	"unstructuredmarkdown": UnstructuredMarkdownLoader,
	# 4. HTML
	"bshtml": BSHTMLLoader,
	# 5. XML
	"unstructuredxml": UnstructuredXMLLoader,
	# 6. All files
	"directory": DirectoryLoader,
	"unstructured": UnstructuredLoader,
	"upstagedocumentparse": UpstageLayoutAnalysisLoader,
}

chunk_modules = {
	# Llama Index
	# Token
	"token": TokenTextSplitter,
	# Sentence
	"sentence": SentenceSplitter,
	# window
	"sentencewindow": SentenceWindowNodeParser,
	# Semantic
	"semantic_llama_index": SemanticSplitterNodeParser,
	"semanticdoublemerging": SemanticDoubleMergingSplitterNodeParser,
	# Simple
	"simplefile": SimpleFileNodeParser,
	# LangChain
	# Token
	"sentencetransformerstoken": SentenceTransformersTokenTextSplitter,
	# Character
	"recursivecharacter": RecursiveCharacterTextSplitter,
	"character": CharacterTextSplitter,
	# Sentence
	"konlpy": KonlpyTextSplitter,
}


def split_by_sentence_kiwi() -> Callable[[str], List[str]]:
	try:
		from kiwipiepy import Kiwi
	except ImportError:
		raise ImportError(
			"You need to install kiwipiepy to use 'ko_kiwi' tokenizer. "
			"Please install kiwipiepy by running 'pip install kiwipiepy'. "
			"Or install Korean version of AutoRAG by running 'pip install AutoRAG[ko]'."
		)
	kiwi = Kiwi()

	def split(text: str) -> List[str]:
		kiwi_result = kiwi.split_into_sents(text)
		sentences = list(map(lambda x: x.text, kiwi_result))

		return sentences

	return split


sentence_splitter_modules = {"kiwi": LazyInit(split_by_sentence_kiwi)}



from typing import List

import pandas as pd

from autorag.nodes.queryexpansion.base import BaseQueryExpansion
from autorag.utils import result_to_dataframe

multi_query_expansion_prompt = """You are an AI language model assistant.
    Your task is to generate 3 different versions of the given user
    question to retrieve relevant documents from a vector  database.
    By generating multiple perspectives on the user question,
    your goal is to help the user overcome some of the limitations
    of distance-based similarity search. Provide these alternative
    questions separated by newlines. Original question: {query}"""


class MultiQueryExpansion(BaseQueryExpansion):
	@result_to_dataframe(["queries"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result, *args, **kwargs)

		# pop prompt from kwargs
		prompt = kwargs.pop("prompt", multi_query_expansion_prompt)
		kwargs.pop("generator_module_type", None)

		expanded_queries = self._pure(queries, prompt, **kwargs)
		return self._check_expanded_query(queries, expanded_queries)

	def _pure(
		self, queries, prompt: str = multi_query_expansion_prompt, **kwargs
	) -> List[List[str]]:
		"""
		Expand a list of queries using a multi-query expansion approach.
		LLM model generate 3 different versions queries for each input query.

		:param queries: List[str], queries to decompose.
		:param prompt: str, prompt to use for multi-query expansion.
			default prompt comes from langchain MultiQueryRetriever default query prompt.
		:return: List[List[str]], list of expansion query.
		"""
		full_prompts = list(map(lambda x: prompt.format(query=x), queries))
		input_df = pd.DataFrame({"prompts": full_prompts})
		result_df = self.generator.pure(previous_result=input_df, **kwargs)
		answers = result_df["generated_texts"].tolist()
		results = list(
			map(lambda x: get_multi_query_expansion(x[0], x[1]), zip(queries, answers))
		)
		return results


def get_multi_query_expansion(query: str, answer: str) -> List[str]:
	try:
		queries = answer.split("\n")
		queries.insert(0, query)
		return queries
	except:
		return [query]



import itertools
import logging
import os
from typing import List, Tuple, Optional

import numpy as np
import pandas as pd
from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.evaluation.metric.util import (
	calculate_l2_distance,
	calculate_inner_product,
	calculate_cosine_similarity,
)
from autorag.nodes.retrieval.base import evenly_distribute_passages, BaseRetrieval
from autorag.utils import (
	validate_corpus_dataset,
	cast_corpus_dataset,
	cast_qa_dataset,
	validate_qa_dataset,
)
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	openai_truncate_by_token,
	flatten_apply,
	result_to_dataframe,
	pop_params,
	fetch_contents,
	empty_cuda_cache,
	convert_inputs_to_list,
	make_batch,
)
from autorag.vectordb import load_vectordb_from_yaml
from autorag.vectordb.base import BaseVectorStore

logger = logging.getLogger("AutoRAG")


class VectorDB(BaseRetrieval):
	def __init__(self, project_dir: str, vectordb: str = "default", **kwargs):
		"""
		Initialize VectorDB retrieval node.

		:param project_dir: The project directory path.
		:param vectordb: The vectordb name.
			You must configure the vectordb name in the config.yaml file.
			If you don't configure, it uses the default vectordb.
		:param kwargs: The optional arguments.
			Not affected in the init method.
		"""
		super().__init__(project_dir)

		vectordb_config_path = os.path.join(self.resources_dir, "vectordb.yaml")
		self.vector_store = load_vectordb_from_yaml(
			vectordb_config_path, vectordb, project_dir
		)

		self.embedding_model = self.vector_store.embedding

	def __del__(self):
		del self.vector_store
		del self.embedding_model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(
		[
			"retrieved_contents_semantic",
			"retrieved_ids_semantic",
			"retrieve_scores_semantic",
		]
	)
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result)
		pure_params = pop_params(self._pure, kwargs)
		ids, scores = self._pure(queries, **pure_params)
		contents = fetch_contents(self.corpus_df, ids)
		return contents, ids, scores

	def _pure(
		self,
		queries: List[List[str]],
		top_k: int,
		embedding_batch: int = 128,
		ids: Optional[List[List[str]]] = None,
	) -> Tuple[List[List[str]], List[List[float]]]:
		"""
		VectorDB retrieval function.
		You have to get a chroma collection that is already ingested.
		You have to get an embedding model that is already used in ingesting.

		:param queries: 2-d list of query strings.
		    Each element of the list is a query strings of each row.
		:param top_k: The number of passages to be retrieved.
		:param embedding_batch: The number of queries to be processed in parallel.
		    This is used to prevent API error at the query embedding.
		    Default is 128.
		:param ids: The optional list of ids that you want to retrieve.
		    You don't need to specify this in the general use cases.
		    Default is None.

		:return: The 2-d list contains a list of passage ids that retrieved from vectordb and 2-d list of its scores.
		    It will be a length of queries. And each element has a length of top_k.
		"""
		# if ids are specified, fetch the ids score from Chroma
		if ids is not None:
			return self.__get_ids_scores(queries, ids, embedding_batch)

		# run async vector_db_pure function
		tasks = [
			vectordb_pure(query_list, top_k, self.vector_store)
			for query_list in queries
		]
		loop = get_event_loop()
		results = loop.run_until_complete(
			process_batch(tasks, batch_size=embedding_batch)
		)
		id_result = list(map(lambda x: x[0], results))
		score_result = list(map(lambda x: x[1], results))
		return id_result, score_result

	def __get_ids_scores(self, queries, ids, embedding_batch: int):
		# truncate queries and embedding execution here.
		openai_embedding_limit = 8000
		if isinstance(self.embedding_model, OpenAIEmbedding):
			queries = list(
				map(
					lambda query_list: openai_truncate_by_token(
						query_list,
						openai_embedding_limit,
						self.embedding_model.model_name,
					),
					queries,
				)
			)

		query_embeddings = flatten_apply(
			run_query_embedding_batch,
			queries,
			embedding_model=self.embedding_model,
			batch_size=embedding_batch,
		)

		loop = get_event_loop()

		async def run_fetch(ids):
			final_result = []
			for id_list in ids:
				if len(id_list) == 0:
					final_result.append([])
				else:
					result = await self.vector_store.fetch(id_list)
					final_result.append(result)
			return final_result

		content_embeddings = loop.run_until_complete(run_fetch(ids))

		score_result = list(
			map(
				lambda query_embedding_list, content_embedding_list: get_id_scores(
					query_embedding_list,
					content_embedding_list,
					similarity_metric=self.vector_store.similarity_metric,
				),
				query_embeddings,
				content_embeddings,
			)
		)
		return ids, score_result


async def vectordb_pure(
	queries: List[str], top_k: int, vectordb: BaseVectorStore
) -> Tuple[List[str], List[float]]:
	"""
	Async VectorDB retrieval function.
	Its usage is for async retrieval of vector_db row by row.

	:param query_embeddings: A list of query embeddings.
	:param top_k: The number of passages to be retrieved.
	:param vectordb: The vector store instance.
	:return: The tuple contains a list of passage ids that are retrieved from vectordb and a list of its scores.
	"""
	id_result, score_result = await vectordb.query(queries=queries, top_k=top_k)

	# Distribute passages evenly
	id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
	# sort id_result and score_result by score
	result = [
		(_id, score)
		for score, _id in sorted(
			zip(score_result, id_result), key=lambda pair: pair[0], reverse=True
		)
	]
	id_result, score_result = zip(*result)
	return list(id_result), list(score_result)


async def filter_exist_ids(
	vectordb: BaseVectorStore,
	corpus_data: pd.DataFrame,
) -> pd.DataFrame:
	corpus_data = cast_corpus_dataset(corpus_data)
	validate_corpus_dataset(corpus_data)
	ids = corpus_data["doc_id"].tolist()

	# Query the collection to check if IDs already exist
	existed_bool_list = await vectordb.is_exist(ids=ids)
	# Assuming 'ids' is the key in the response
	new_passage = corpus_data[~pd.Series(existed_bool_list)]
	return new_passage


async def filter_exist_ids_from_retrieval_gt(
	vectordb: BaseVectorStore,
	qa_data: pd.DataFrame,
	corpus_data: pd.DataFrame,
) -> pd.DataFrame:
	qa_data = cast_qa_dataset(qa_data)
	validate_qa_dataset(qa_data)
	corpus_data = cast_corpus_dataset(corpus_data)
	validate_corpus_dataset(corpus_data)
	retrieval_gt = (
		qa_data["retrieval_gt"]
		.apply(lambda x: list(itertools.chain.from_iterable(x)))
		.tolist()
	)
	retrieval_gt = list(itertools.chain.from_iterable(retrieval_gt))
	retrieval_gt = list(set(retrieval_gt))

	existed_bool_list = await vectordb.is_exist(ids=retrieval_gt)
	add_ids = []
	for ret_gt, is_exist in zip(retrieval_gt, existed_bool_list):
		if not is_exist:
			add_ids.append(ret_gt)
	new_passage = corpus_data[corpus_data["doc_id"].isin(add_ids)]
	return new_passage


async def vectordb_ingest_api(
	vectordb: BaseVectorStore,
	corpus_data: pd.DataFrame,
):
	"""
	Ingest given corpus data to the vectordb.
	It truncates corpus content when the embedding model is OpenAIEmbedding to the 8000 tokens.
	Plus, when the corpus content is empty (whitespace), it will be ignored.
	And if there is a document id that already exists in the collection, it will be ignored.

	:param vectordb: A vector stores instance that you want to ingest.
	:param corpus_data: The corpus data that contains doc_id and contents columns.
	"""
	embedding_batch = vectordb.embedding_batch
	if not corpus_data.empty:
		new_contents = corpus_data["contents"].tolist()
		new_ids = corpus_data["doc_id"].tolist()
		content_batches = make_batch(new_contents, embedding_batch)
		id_batches = make_batch(new_ids, embedding_batch)
		for content_batch, id_batch in zip(content_batches, id_batches):
			await vectordb.add(ids=id_batch, texts=content_batch)


def vectordb_ingest_huggingface(
	vectordb: BaseVectorStore,
	corpus_data: pd.DataFrame,
):
	"""
	Ingest given corpus data to the vectordb using local model.
	When the corpus content is empty (whitespace), it will be ignored.
	And if there is a document id that already exists in the collection, it will be ignored.

	:param vectordb: A vector stores instance that you want to ingest.
	:param corpus_data: The corpus data that contains doc_id and contents columns.
	"""
	embedding_batch_size = vectordb.embedding_batch
	embedding_model = vectordb.embedding._model
	if corpus_data.empty:
		logger.warning("The corpus data is empty. Nothing to ingest.")
		return
	new_contents = corpus_data["contents"].tolist()
	new_ids = corpus_data["doc_id"].tolist()
	logger.info("Start embedding corpus data with huggingface model.")
	embeddings = embedding_model.encode(
		new_contents,
		batch_size=embedding_batch_size,
		normalize_embeddings=vectordb.embedding.normalize,
		show_progrss_bar=True,
	)
	vectordb.add_embedding(new_ids, embeddings)
	logger.info("Finish embedding & ingesting corpus data with huggingface model.")


def run_query_embedding_batch(
	queries: List[str], embedding_model: BaseEmbedding, batch_size: int
) -> List[List[float]]:
	result = []
	for i in range(0, len(queries), batch_size):
		batch = queries[i : i + batch_size]
		embeddings = embedding_model.get_text_embedding_batch(batch)
		result.extend(embeddings)
	return result


@convert_inputs_to_list
def get_id_scores(  # To find the uncalculated score when fuse the scores for the hybrid retrieval
	query_embeddings: List[
		List[float]
	],  # `queries` is input. This is one user input query.
	content_embeddings: List[List[float]],
	similarity_metric: str,
) -> List[
	float
]:  # The most high scores among each query. The length of a result is the same as the contents length.
	"""
	Calculate the highest similarity scores between query embeddings and content embeddings.

	:param query_embeddings: A list of lists containing query embeddings.
	:param content_embeddings: A list of lists containing content embeddings.
	:param similarity_metric: The similarity metric to use ('l2', 'ip', or 'cosine').
	:return: A list of the highest similarity scores for each content embedding.
	"""
	metric_func_dict = {
		"l2": lambda x, y: 1 - calculate_l2_distance(x, y),
		"ip": calculate_inner_product,
		"cosine": calculate_cosine_similarity,
	}
	metric_func = metric_func_dict[similarity_metric]

	result = []
	for content_embedding in content_embeddings:
		scores = []
		for query_embedding in query_embeddings:
			scores.append(
				metric_func(np.array(query_embedding), np.array(content_embedding))
			)
		result.append(max(scores))
	return result



import os
from typing import List, Tuple
import pandas as pd
import voyageai

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import result_to_dataframe, get_event_loop, process_batch


class VoyageAIReranker(BasePassageReranker):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir)
		api_key = kwargs.pop("api_key", None)
		api_key = os.getenv("VOYAGE_API_KEY", None) if api_key is None else api_key
		if api_key is None:
			raise KeyError(
				"Please set the API key for VoyageAI rerank in the environment variable VOYAGE_API_KEY "
				"or directly set it on the config YAML file."
			)

		self.voyage_client = voyageai.AsyncClient(api_key=api_key)

	def __del__(self):
		del self.voyage_client
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 8)
		model = kwargs.pop("model", "rerank-2")
		truncation = kwargs.pop("truncation", True)
		return self._pure(queries, contents, ids, top_k, model, batch, truncation)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		model: str = "rerank-2",
		batch: int = 8,
		truncation: bool = True,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with VoyageAI rerank models.
		You can get the API key from https://docs.voyageai.com/docs/api-key-and-installation and set it in the environment variable VOYAGE_API_KEY.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param model: The model name for VoyageAI rerank.
		    You can choose between "rerank-2" and "rerank-2-lite".
		    Default is "rerank-2".
		:param batch: The number of queries to be processed in a batch
		:param truncation: Whether to truncate the input to satisfy the 'context length limit' on the query and the documents.
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		tasks = [
			voyageai_rerank_pure(
				self.voyage_client, model, query, contents, ids, top_k, truncation
			)
			for query, contents, ids in zip(queries, contents_list, ids_list)
		]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch))

		content_result, id_result, score_result = zip(*results)

		return list(content_result), list(id_result), list(score_result)


async def voyageai_rerank_pure(
	voyage_client: voyageai.AsyncClient,
	model: str,
	query: str,
	documents: List[str],
	ids: List[str],
	top_k: int,
	truncation: bool = True,
) -> Tuple[List[str], List[str], List[float]]:
	"""
	Rerank a list of contents with VoyageAI rerank models.

	:param voyage_client: The Voyage Client to use for reranking
	:param model: The model name for VoyageAI rerank
	:param query: The query to use for reranking
	:param documents: The list of contents to rerank
	:param ids: The list of ids corresponding to the documents
	:param top_k: The number of passages to be retrieved
	:param truncation: Whether to truncate the input to satisfy the 'context length limit' on the query and the documents.
	:return: Tuple of lists containing the reranked contents, ids, and scores
	"""
	rerank_results = await voyage_client.rerank(
		model=model,
		query=query,
		documents=documents,
		top_k=top_k,
		truncation=truncation,
	)
	reranked_scores: List[float] = list(
		map(lambda x: x.relevance_score, rerank_results.results)
	)
	indices = list(map(lambda x: x.index, rerank_results.results))
	reranked_contents: List[str] = list(map(lambda i: documents[i], indices))
	reranked_ids: List[str] = list(map(lambda i: ids[i], indices))
	return reranked_contents, reranked_ids, reranked_scores



from typing import List

import pandas as pd

from autorag.nodes.promptmaker.base import BasePromptMaker
from autorag.utils import result_to_dataframe


class Fstring(BasePromptMaker):
	@result_to_dataframe(["prompts"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		query, retrieved_contents, prompt = self.cast_to_run(
			previous_result, *args, **kwargs
		)
		return self._pure(prompt, query, retrieved_contents)

	def _pure(
		self, prompt: str, queries: List[str], retrieved_contents: List[List[str]]
	) -> List[str]:
		"""
		Make a prompt using f-string from a query and retrieved_contents.
		You must type a prompt or prompt list at a config YAML file like this:

		.. Code:: yaml
		nodes:
		- node_type: prompt_maker
		  modules:
		  - module_type: fstring
			prompt: [Answer this question: {query} \n\n {retrieved_contents},
			Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

		:param prompt: A prompt string.
		:param queries: List of query strings.
		:param retrieved_contents: List of retrieved contents.
		:return: Prompts that are made by f-string.
		"""

		def fstring_row(
			_prompt: str, _query: str, _retrieved_contents: List[str]
		) -> str:
			contents_str = "\n\n".join(_retrieved_contents)
			return _prompt.format(query=_query, retrieved_contents=contents_str)

		return list(
			map(
				lambda x: fstring_row(prompt, x[0], x[1]),
				zip(queries, retrieved_contents),
			)
		)



from itertools import chain
from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	pop_params,
	empty_cuda_cache,
)

prediction_tokens = {
	"castorini/monot5-base-msmarco": ["▁false", "▁true"],
	"castorini/monot5-base-msmarco-10k": ["▁false", "▁true"],
	"castorini/monot5-large-msmarco": ["▁false", "▁true"],
	"castorini/monot5-large-msmarco-10k": ["▁false", "▁true"],
	"castorini/monot5-base-med-msmarco": ["▁false", "▁true"],
	"castorini/monot5-3b-med-msmarco": ["▁false", "▁true"],
	"castorini/monot5-3b-msmarco-10k": ["▁false", "▁true"],
	"unicamp-dl/mt5-base-en-msmarco": ["▁no", "▁yes"],
	"unicamp-dl/ptt5-base-pt-msmarco-10k-v2": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-pt-msmarco-100k-v2": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2": ["▁não", "▁sim"],
	"unicamp-dl/mt5-base-en-pt-msmarco-v2": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-mmarco-v2": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-en-pt-msmarco-v1": ["▁no", "▁yes"],
	"unicamp-dl/mt5-base-mmarco-v1": ["▁no", "▁yes"],
	"unicamp-dl/ptt5-base-pt-msmarco-10k-v1": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-pt-msmarco-100k-v1": ["▁não", "▁sim"],
	"unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1": ["▁não", "▁sim"],
	"unicamp-dl/mt5-3B-mmarco-en-pt": ["▁", "▁true"],
	"unicamp-dl/mt5-13b-mmarco-100k": ["▁", "▁true"],
}


class MonoT5(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model_name: str = "castorini/monot5-3b-msmarco-10k",
		*args,
		**kwargs,
	):
		"""
		Initialize the MonoT5 reranker.

		:param project_dir: The project directory
		:param model_name: The name of the MonoT5 model to use for reranking
			Note: default model name is 'castorini/monot5-3b-msmarco-10k'
				If there is a '/' in the model name parameter,
				when we create the file to store the results, the path will be twisted because of the '/'.
				Therefore, it will be received as '_' instead of '/'.
		:param kwargs: The extra arguments for the MonoT5 reranker
		"""
		super().__init__(project_dir)
		try:
			import torch
			from transformers import T5Tokenizer, T5ForConditionalGeneration
		except ImportError:
			raise ImportError("For using MonoT5 Reranker, please install torch first.")
		# replace '_' to '/'
		if "_" in model_name:
			model_name = model_name.replace("_", "/")
		# Load the tokenizer and model from the pre-trained MonoT5 model
		self.tokenizer = T5Tokenizer.from_pretrained(model_name)
		model_params = pop_params(T5ForConditionalGeneration.from_pretrained, kwargs)
		self.model = T5ForConditionalGeneration.from_pretrained(
			model_name, **model_params
		).eval()

		# Determine the device to run the model on (GPU if available, otherwise CPU)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		self.model.to(self.device)

		token_false, token_true = prediction_tokens[model_name]
		self.token_false_id = self.tokenizer.convert_tokens_to_ids(token_false)
		self.token_true_id = self.tokenizer.convert_tokens_to_ids(token_true)

	def __del__(self):
		del self.model
		del self.tokenizer
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.get("top_k", 3)
		batch = kwargs.get("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using MonoT5.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved

		:param batch: The number of queries to be processed in a batch
		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		# Retrieve the tokens used by the model to represent false and true predictions

		nested_list = [
			list(map(lambda x: [f"Query: {query} Document: {x}"], content_list))
			for query, content_list in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			monot5_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
			device=self.device,
			token_false_id=self.token_false_id,
			token_true_id=self.token_true_id,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def monot5_run_model(
	input_texts,
	model,
	batch_size: int,
	tokenizer,
	device,
	token_false_id,
	token_true_id,
):
	try:
		import torch
	except ImportError:
		raise ImportError("For using MonoT5 Reranker, please install torch first.")
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in batch_input_texts:
		flattened_batch_texts = list(chain.from_iterable(batch_texts))
		input_encodings = tokenizer(
			flattened_batch_texts,
			padding=True,
			truncation=True,
			max_length=512,
			return_tensors="pt",
		).to(device)
		with torch.no_grad():
			outputs = model.generate(
				input_ids=input_encodings["input_ids"],
				attention_mask=input_encodings["attention_mask"],
				output_scores=True,
				return_dict_in_generate=True,
			)

		# Extract logits for the 'false' and 'true' tokens from the model's output
		logits = outputs.scores[-1][:, [token_false_id, token_true_id]]
		# Calculate the softmax probability of the 'true' token
		probs = torch.nn.functional.softmax(logits, dim=-1)[:, 1]
		results.extend(probs.tolist())
	return results



import os
from typing import List, Tuple

import pandas as pd
from mixedbread_ai.client import AsyncMixedbreadAI

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	result_to_dataframe,
	get_event_loop,
	process_batch,
	pop_params,
)


class MixedbreadAIReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		*args,
		**kwargs,
	):
		"""
		Initialize mixedbread-ai rerank node.

		:param project_dir: The project directory path.
		:param api_key: The API key for MixedbreadAI rerank.
		    You can set it in the environment variable MXBAI_API_KEY.
		    Or, you can directly set it on the config YAML file using this parameter.
		    Default is env variable "MXBAI_API_KEY".
		:param kwargs: Extra arguments that are not affected
		"""
		super().__init__(project_dir)
		api_key = kwargs.pop("api_key", None)
		api_key = os.getenv("MXBAI_API_KEY", None) if api_key is None else api_key
		if api_key is None:
			raise KeyError(
				"Please set the API key for Mixedbread AI rerank in the environment variable MXBAI_API_KEY "
				"or directly set it on the config YAML file."
			)
		self.client = AsyncMixedbreadAI(api_key=api_key)

	def __del__(self):
		del self.client
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 8)
		model = kwargs.pop("model", "mixedbread-ai/mxbai-rerank-large-v1")
		rerank_params = pop_params(self.client.reranking, kwargs)
		return self._pure(queries, contents, ids, top_k, model, batch, **rerank_params)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		model: str = "mixedbread-ai/mxbai-rerank-large-v1",
		batch: int = 8,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with mixedbread-ai rerank models.
		You can get the API key from https://www.mixedbread.ai/api-reference#quick-start-guide and set it in the environment variable MXBAI_API_KEY.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param model: The model name for mixedbread-ai rerank.
			You can choose between "mixedbread-ai/mxbai-rerank-large-v1", "mixedbread-ai/mxbai-rerank-base-v1" and "mixedbread-ai/mxbai-rerank-xsmall-v1".
			Default is "mixedbread-ai/mxbai-rerank-large-v1".
		:param batch: The number of queries to be processed in a batch
		        :return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		tasks = [
			mixedbreadai_rerank_pure(
				self.client, query, contents, ids, top_k=top_k, model=model
			)
			for query, contents, ids in zip(queries, contents_list, ids_list)
		]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch))

		content_result, id_result, score_result = zip(*results)

		return list(content_result), list(id_result), list(score_result)


async def mixedbreadai_rerank_pure(
	client: AsyncMixedbreadAI,
	query: str,
	documents: List[str],
	ids: List[str],
	top_k: int,
	model: str = "mixedbread-ai/mxbai-rerank-large-v1",
) -> Tuple[List[str], List[str], List[float]]:
	"""
	Rerank a list of contents with mixedbread-ai rerank models.

	:param client: The mixedbread-ai client to use for reranking
	:param query: The query to use for reranking
	:param documents: The list of contents to rerank
	:param ids: The list of ids corresponding to the documents
	:param top_k: The number of passages to be retrieved
	:param model: The model name for mixedbread-ai rerank.
	    You can choose between "mixedbread-ai/mxbai-rerank-large-v1" and "mixedbread-ai/mxbai-rerank-base-v1".
	    Default is "mixedbread-ai/mxbai-rerank-large-v1".
	:return: Tuple of lists containing the reranked contents, ids, and scores
	"""

	results = await client.reranking(
		query=query,
		input=documents,
		top_k=top_k,
		model=model,
	)
	reranked_scores: List[float] = list(map(lambda x: x.score, results.data))
	reranked_scores_float = list(map(float, reranked_scores))
	indices = list(map(lambda x: x.index, results.data))
	reranked_contents = list(map(lambda x: documents[x], indices))
	reranked_ids: List[str] = list(map(lambda i: ids[i], indices))
	return reranked_contents, reranked_ids, reranked_scores_float



import abc
import logging
import os
from typing import List, Union, Tuple

import pandas as pd

from autorag.schema import BaseModule
from autorag.support import get_support_modules
from autorag.utils import fetch_contents, result_to_dataframe, validate_qa_dataset
from autorag.utils.util import pop_params

logger = logging.getLogger("AutoRAG")


class BaseRetrieval(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: str, *args, **kwargs):
		logger.info(f"Initialize retrieval node - {self.__class__.__name__}")

		self.resources_dir = os.path.join(project_dir, "resources")
		data_dir = os.path.join(project_dir, "data")
		# fetch data from corpus_data
		self.corpus_df = pd.read_parquet(
			os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
		)

	def __del__(self):
		logger.info(f"Deleting retrieval node - {self.__class__.__name__} module...")

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(f"Running retrieval node - {self.__class__.__name__} module...")
		validate_qa_dataset(previous_result)
		# find queries columns & type cast queries
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		if "queries" not in previous_result.columns:
			previous_result["queries"] = previous_result["query"]
		previous_result.loc[:, "queries"] = previous_result["queries"].apply(
			cast_queries
		)
		queries = previous_result["queries"].tolist()
		return queries


class HybridRetrieval(BaseRetrieval, metaclass=abc.ABCMeta):
	def __init__(
		self, project_dir: str, target_modules, target_module_params, *args, **kwargs
	):
		super().__init__(project_dir)
		self.target_modules = list(
			map(
				lambda x, y: get_support_modules(x)(
					**y,
					project_dir=project_dir,
				),
				target_modules,
				target_module_params,
			)
		)
		self.target_module_params = target_module_params

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		result_dfs: List[pd.DataFrame] = list(
			map(
				lambda x, y: x.pure(
					**y,
					previous_result=previous_result,
				),
				self.target_modules,
				self.target_module_params,
			)
		)
		ids = tuple(
			map(lambda df: df["retrieved_ids"].apply(list).tolist(), result_dfs)
		)
		scores = tuple(
			map(
				lambda df: df["retrieve_scores"].apply(list).tolist(),
				result_dfs,
			)
		)

		_pure_params = pop_params(self._pure, kwargs)
		if "ids" in _pure_params or "scores" in _pure_params:
			raise ValueError(
				"With specifying ids or scores, you must use HybridRRF.run_evaluator instead."
			)
		ids, scores = self._pure(ids=ids, scores=scores, **_pure_params)
		contents = fetch_contents(self.corpus_df, ids)
		return contents, ids, scores


def cast_queries(queries: Union[str, List[str]]) -> List[str]:
	if isinstance(queries, str):
		return [queries]
	elif isinstance(queries, List):
		return queries
	else:
		raise ValueError(f"queries must be str or list, but got {type(queries)}")


def evenly_distribute_passages(
	ids: List[List[str]], scores: List[List[float]], top_k: int
) -> Tuple[List[str], List[float]]:
	assert len(ids) == len(scores), "ids and scores must have same length."
	query_cnt = len(ids)
	avg_len = top_k // query_cnt
	remainder = top_k % query_cnt

	new_ids = []
	new_scores = []
	for i in range(query_cnt):
		if i < remainder:
			new_ids.extend(ids[i][: avg_len + 1])
			new_scores.extend(scores[i][: avg_len + 1])
		else:
			new_ids.extend(ids[i][:avg_len])
			new_scores.extend(scores[i][:avg_len])

	return new_ids, new_scores


def get_bm25_pkl_name(bm25_tokenizer: str):
	bm25_tokenizer = bm25_tokenizer.replace("/", "")
	return f"bm25_{bm25_tokenizer}.pkl"



import abc
import functools
import logging
from pathlib import Path
from typing import Union, Tuple, List

import pandas as pd
from llama_index.core.output_parsers import PydanticOutputParser

from autorag import generator_models
from autorag.schema import BaseModule
from autorag.utils import result_to_dataframe

logger = logging.getLogger("AutoRAG")


class BaseGenerator(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: str, llm: str, *args, **kwargs):
		logger.info(f"Initialize generator node - {self.__class__.__name__}")
		self.llm = llm

	def __del__(self):
		logger.info(f"Deleting generator module - {self.__class__.__name__}")

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(f"Running generator node - {self.__class__.__name__} module...")
		assert (
			"prompts" in previous_result.columns
		), "previous_result must contain prompts column."
		prompts = previous_result["prompts"].tolist()
		return prompts

	def structured_output(self, prompts: List[str], output_cls):
		response, _, _ = self._pure(prompts)
		parser = PydanticOutputParser(output_cls)
		result = []
		for res in response:
			try:
				result.append(parser.parse(res))
			except Exception as e:
				logger.warning(
					f"Error parsing response: {e} \nSo returning None instead in this case."
				)
				result.append(None)
		return result

	@abc.abstractmethod
	async def astream(self, prompt: str, **kwargs):
		pass

	@abc.abstractmethod
	def stream(self, prompt: str, **kwargs):
		pass


def generator_node(func):
	@functools.wraps(func)
	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def wrapper(
		project_dir: Union[str, Path], previous_result: pd.DataFrame, llm: str, **kwargs
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		This decorator makes a generator module to be a node.
		It automatically extracts prompts from previous_result and runs the generator function.
		Plus, it retrieves the llm instance from autorag.generator_models.

		:param project_dir: The project directory.
		:param previous_result: The previous result that contains prompts,
		:param llm: The llm name that you want to use.
		:param kwargs: The extra parameters for initializing the llm instance.
		:return: Pandas dataframe that contains generated texts, generated tokens, and generated log probs.
		    Each column is "generated_texts", "generated_tokens", and "generated_log_probs".
		"""
		logger.info(f"Running generator node - {func.__name__} module...")
		assert (
			"prompts" in previous_result.columns
		), "previous_result must contain prompts column."
		prompts = previous_result["prompts"].tolist()
		if func.__name__ == "llama_index_llm":
			if llm not in generator_models:
				raise ValueError(
					f"{llm} is not a valid llm name. Please check the llm name."
					"You can check valid llm names from autorag.generator_models."
				)
			batch = kwargs.pop("batch", 16)
			if llm == "huggingfacellm":
				model_name = kwargs.pop("model", None)
				if model_name is not None:
					kwargs["model_name"] = model_name
				else:
					if "model_name" not in kwargs.keys():
						raise ValueError(
							"`model` or `model_name` parameter must be provided for using huggingfacellm."
						)
				kwargs["tokenizer_name"] = kwargs["model_name"]
			llm_instance = generator_models[llm](**kwargs)
			result = func(prompts=prompts, llm=llm_instance, batch=batch)
			del llm_instance
			return result
		else:
			return func(prompts=prompts, llm=llm, **kwargs)

	return wrapper



import logging
import uuid

import pandas as pd

from autorag.deploy.base import BaseRunner

import gradio as gr


logger = logging.getLogger("AutoRAG")


class GradioRunner(BaseRunner):
	def run_web(
		self,
		server_name: str = "0.0.0.0",
		server_port: int = 7680,
		share: bool = False,
		**kwargs,
	):
		"""
		Run web interface to interact pipeline.
		You can access the web interface at `http://server_name:server_port` in your browser

		:param server_name: The host of the web. Default is 0.0.0.0.
		:param server_port: The port of the web. Default is 7680.
		:param share: Whether to create a publicly shareable link. Default is False.
		:param kwargs: Other arguments for gr.ChatInterface.launch.
		"""

		logger.info(f"Run web interface at http://{server_name}:{server_port}")

		def get_response(message, _):
			return self.run(message)

		gr.ChatInterface(
			get_response, title="📚 AutoRAG", retry_btn=None, undo_btn=None
		).launch(
			server_name=server_name, server_port=server_port, share=share, **kwargs
		)

	def run(self, query: str, result_column: str = "generated_texts"):
		"""
		Run the pipeline with query.
		The loaded pipeline must start with a single query,
		so the first module of the pipeline must be `query_expansion` or `retrieval` module.

		:param query: The query of the user.
		:param result_column: The result column name for the answer.
		    Default is `generated_texts`, which is the output of the `generation` module.
		:return: The result of the pipeline.
		"""
		previous_result = pd.DataFrame(
			{
				"qid": str(uuid.uuid4()),
				"query": [query],
				"retrieval_gt": [[]],
				"generation_gt": [""],
			}
		)  # pseudo qa data for execution
		for module_instance, module_param in zip(
			self.module_instances, self.module_params
		):
			new_result = module_instance.pure(
				previous_result=previous_result, **module_param
			)
			duplicated_columns = previous_result.columns.intersection(
				new_result.columns
			)
			drop_previous_result = previous_result.drop(columns=duplicated_columns)
			previous_result = pd.concat([drop_previous_result, new_result], axis=1)

		return previous_result[result_column].tolist()[0]



import logging
import os
import shutil
from typing import Optional

import pandas as pd

from autorag.data.chunk.run import run_chunker
from autorag.data.utils.util import load_yaml, get_param_combinations

logger = logging.getLogger("AutoRAG")


class Chunker:
	def __init__(self, raw_df: pd.DataFrame, project_dir: Optional[str] = None):
		self.parsed_raw = raw_df
		self.project_dir = project_dir if project_dir is not None else os.getcwd()

	@classmethod
	def from_parquet(
		cls, parsed_data_path: str, project_dir: Optional[str] = None
	) -> "Chunker":
		if not os.path.exists(parsed_data_path):
			raise ValueError(f"parsed_data_path {parsed_data_path} does not exist.")
		if not parsed_data_path.endswith("parquet"):
			raise ValueError(
				f"parsed_data_path {parsed_data_path} is not a parquet file."
			)
		parsed_result = pd.read_parquet(parsed_data_path, engine="pyarrow")
		return cls(parsed_result, project_dir)

	def start_chunking(self, yaml_path: str):
		if not os.path.exists(self.project_dir):
			os.makedirs(self.project_dir)

		# Copy YAML file to the trial directory
		shutil.copy(yaml_path, os.path.join(self.project_dir, "chunk_config.yaml"))

		# load yaml file
		modules = load_yaml(yaml_path)

		input_modules, input_params = get_param_combinations(modules)

		logger.info("Chunking Start...")
		run_chunker(
			modules=input_modules,
			module_params=input_params,
			parsed_result=self.parsed_raw,
			project_dir=self.project_dir,
		)
		logger.info("Chunking Done!")



import os
from typing import Callable, List, Dict
import pandas as pd

from autorag.strategy import measure_speed


def run_chunker(
	modules: List[Callable],
	module_params: List[Dict],
	parsed_result: pd.DataFrame,
	project_dir: str,
):
	results, execution_times = zip(
		*map(
			lambda x: measure_speed(x[0], parsed_result=parsed_result, **x[1]),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# save results to parquet files
	filepaths = list(
		map(lambda x: os.path.join(project_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths)))
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
		}
	)
	summary_df.to_csv(os.path.join(project_dir, "summary.csv"), index=False)
	return summary_df



import functools
import warnings
from typing import List, Callable, Union, Dict

import pandas as pd

from autorag.evaluation.metric.generation import (
	bleu,
	meteor,
	rouge,
	sem_score,
	g_eval,
	bert_score,
	deepeval_faithfulness,
)
from autorag.evaluation.util import cast_metrics
from autorag.schema.metricinput import MetricInput

GENERATION_METRIC_FUNC_DICT = {
	func.__name__: func
	for func in [
		bleu,
		meteor,
		rouge,
		sem_score,
		g_eval,
		bert_score,
		deepeval_faithfulness,
	]
}


def evaluate_generation(
	metric_inputs: List[MetricInput], metrics: Union[List[str], List[Dict]]
):
	def decorator_evaluate_generation(func: Callable):
		@functools.wraps(func)
		def wrapper(*args, **kwargs) -> pd.DataFrame:
			generation_result = func(*args, **kwargs)
			if type(generation_result) is tuple:
				assert (
					type(generation_result[0]) is list
					and type(generation_result[0][0]) is str
				), "Input func must return string list as generated answer at the first return value."
				generated_str = generation_result[0]
			elif type(generation_result) is list:
				assert (
					type(generation_result[0]) is str
				), "Input func must return string list as generated answer at the first return value."
				generated_str = generation_result
			else:
				raise ValueError(
					"Input func must return string list as generated answer at the first return value."
				)
			for metric_input, generated_text in zip(metric_inputs, generated_str):
				metric_input.generated_texts = generated_text

			metric_scores = {}
			metric_names, metric_params = cast_metrics(metrics)

			for metric_name, metric_param in zip(metric_names, metric_params):
				if metric_name not in GENERATION_METRIC_FUNC_DICT:
					warnings.warn(
						f"metric {metric_name} is not in supported metrics: {GENERATION_METRIC_FUNC_DICT.keys()}"
						f"{metric_name} will be ignored."
					)
				else:
					metric_scores[metric_name] = GENERATION_METRIC_FUNC_DICT[
						metric_name
					](
						metric_inputs=metric_inputs,
						**metric_param,
					)

			metric_result_df = pd.DataFrame(metric_scores)
			execution_result_df = pd.DataFrame({"generated_texts": generated_str})
			if type(generation_result) is tuple:
				execution_result_df["generated_tokens"] = generation_result[1]
				execution_result_df["generated_log_probs"] = generation_result[2]

			result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
			return result_df

		return wrapper

	return decorator_evaluate_generation



from typing import List

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils import result_to_dataframe


class PassReranker(BasePassageReranker):
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		top_k = kwargs.pop("top_k")

		_, contents_list, scores_list, ids_list = self.cast_to_run(previous_result)
		return self._pure(contents_list, scores_list, ids_list, top_k)

	def _pure(
		self,
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		top_k: int,
	):
		"""
		Do not perform reranking.
		Return the given top-k passages as is.
		"""
		contents_list = list(map(lambda x: x[:top_k], contents_list))
		scores_list = list(map(lambda x: x[:top_k], scores_list))
		ids_list = list(map(lambda x: x[:top_k], ids_list))
		return contents_list, ids_list, scores_list
